import sys
import os

from intelliw.utils import iuap_request
from intelliw.utils.storage_service import StorageService


def download(url, path, dry: bool = True):
    try:
        if not dry:
            raise Exception()
        iuap_request.download(url=url, output_path=path)
    except Exception:
        StorageService(url, "download").download(path)
    return


if __name__ == '__main__':
    file_url = sys.argv[1]
    save_path = sys.argv[2]
    download(file_url, save_path)



'''
Author: Hexu
Date: 2022-09-27 10:01:23
LastEditors: Hexu
LastEditTime: 2022-09-28 15:09:40
FilePath: /iw-algo-fx/intelliw/utils/dataprocess/__init__.py
Description: 
'''
from .engine import Engine


#!/usr/bin/env python
# coding: utf-8

import json
import threading
import traceback
from intelliw.utils.iuap_request import post
from intelliw.utils import get_json_encoder
from intelliw.utils.logger import _get_framework_logger
from intelliw.config import config
from intelliw.utils.global_val import gl

logger = _get_framework_logger()


class Recorder:
    # 单例锁
    _instance_lock = threading.Lock()

    def __new__(cls, *args, **kwargs):
        """ 单例，防止调用生成更多环境变量dict """
        if not hasattr(Recorder, "_instance"):
            with Recorder._instance_lock:
                if not hasattr(Recorder, "_instance"):
                    Recorder._instance = object.__new__(cls)
                    gl.recorder = Recorder._instance
        return Recorder._instance

    def __init__(self, addr: str):
        # 上报地址
        assert isinstance(addr, (str, type(None))), "Recorder Addr must string"
        self.addr = addr
        self.seq = 0

    def report(self, msg, stdout=True):
        """
        上报
        """
        self.seq += 1
        trace_id = config.SERVICE_ID + '_p' + str(self.seq)
        if hasattr(type(msg), '__str__'):
            data = str(msg)
        else:
            data = json.dumps(msg, ensure_ascii=False,
                              cls=get_json_encoder())
        if self.addr is not None:
            logger.info('[report %s] start, Request: %s', trace_id, msg)
            try:
                headers = {'Content-Type': 'application/json',
                           'X-traceId': trace_id}
                response = post(url=self.addr, headers=headers,
                                data=data.encode('utf-8'))
                response.raise_for_status()
                if stdout:
                    logger.info(
                        '[report %s] success, Response: %s', trace_id, response.body)
            except Exception as e:
                stack_info = traceback.format_exc()
                logger.error(
                    "[report %s] failed, url: [%s], exception: [%s], stack:\n%s", trace_id, self.addr, e, stack_info)
        else:
            logger.info('[report local]Request: %s', msg)



from abc import ABC, abstractmethod
from typing import Dict, Any, Optional


class BaseModelManager(ABC):
    """
    模型管理器基础抽象类
    定义模型管理器的核心接口
    """

    @abstractmethod
    def get_model_version(self) -> str:
        """获取当前模型版本"""
        pass

    @abstractmethod
    def get_active_model(self) -> Any:
        """获取当前激活的模型"""
        pass

    @abstractmethod
    def load_model(self, model_path: str, version: str) -> bool:
        """
        加载指定路径的模型

        :param model_path: 模型路径
        :param version: 模型版本
        :return: 是否加载成功
        """
        pass

    @abstractmethod
    def check_for_updates(self) -> bool:
        """
        检查模型更新

        :return: 是否有更新
        """
        pass

    @abstractmethod
    def start(self) -> None:
        """启动模型管理器"""
        pass

    @abstractmethod
    def stop(self) -> None:
        """停止模型管理器"""
        pass

    @abstractmethod
    def set_config(self, config: Dict[str, Any]) -> None:
        """
        设置配置

        :param config: 配置字典
        """
        pass

    @abstractmethod
    def get_config(self) -> Dict[str, Any]:
        """获取配置"""
        pass


class ModelLoader(ABC):
    """
    模型加载器基础抽象类
    定义不同类型模型的加载接口
    """

    @abstractmethod
    def load(self, model_path: str, load_func) -> Any:
        """
        加载模型

        :param model_path: 模型路径
        :return: 加载的模型
        """
        pass

    @abstractmethod
    def validate(self, model: Any) -> bool:
        """
        验证模型有效性

        :param model: 模型对象
        :return: 是否有效
        """
        pass


class ConfigManager(ABC):
    """
    配置管理器基础抽象类
    """

    @abstractmethod
    def load_config(self, config_path: Optional[str] = None) -> Dict[str, Any]:
        """
        加载配置

        :param config_path: 配置文件路径
        :return: 配置字典
        """
        pass

    @abstractmethod
    def get(self, key: str, default: Any = None) -> Any:
        """
        获取配置项

        :param key: 配置键
        :param default: 默认值
        :return: 配置值
        """
        pass

    @abstractmethod
    def set(self, key: str, value: Any) -> None:
        """
        设置配置项

        :param key: 配置键
        :param value: 配置值
        """
        pass


'''
Author: Hexu
Date: 2022-09-26 13:34:19
LastEditors: Hexu
LastEditTime: 2022-09-26 13:46:59
FilePath: /iw-algo-fx/intelliw/utils/database/proxy.py
Description:

代理模式+装饰器模式实现AOP（面向切面编程）与IOC（控制反转）

目的：非侵入式地给SQLAlchemy的核心类：Session与Query加入横切面，在横切面中可以方便地加入新功能
比如可以扩展增加日志/安全/缓存和事务管理等功能，却不会影响到SQLAlchemy的原生功能。
'''

from sqlalchemy.orm.session import Session


class QueryProxy(object):

    def __init__(self, query):
        self.query = query

    def __getattr__(self, item):
        return getattr(self.query, item)

    def __call__(self, *args, **kwargs):
        return self.query.__call__(*args, **kwargs)


class SessionProxy(object):

    def __init__(self, session, query_proxy_cls=QueryProxy):
        self.session = session
        self.query_proxy_cls = query_proxy_cls

    def __getattr__(self, item):
        if item == 'query':
            query_init_func = self.session.query
            decorated_query_init_func = decorate_query_init_func(
                query_init_func, self.query_proxy_cls)
            return decorated_query_init_func
        else:
            return getattr(self.session, item)


def make_session_proxy(session, session_proxy_cls=SessionProxy, query_proxy_cls=QueryProxy) -> Session:
    """
    这里做了一个Hack。通过代理模式在Session外面wrap一层，
    也能够告诉IDE，返回的对象仍然是Session类，这样的话能够在代理的同时
    还能通过点符号来获取函数补全功能。
    """
    session_proxy: Session = session_proxy_cls(session, query_proxy_cls)
    return session_proxy


def decorate_query_init_func(query_init_func, query_proxy_cls):
    def make_query_proxy(*args, **kwargs):
        query = query_init_func(*args, **kwargs)
        query_proxy = query_proxy_cls(query)
        return query_proxy

    return make_query_proxy



'''
Author: Hexu
Date: 2022-03-14 09:53:59
LastEditors: Hexu
LastEditTime: 2023-05-24 15:59:01
FilePath: /iw-algo-fx/intelliw/core/infer.py
Description: Infer core
'''
import os

# coding: utf-8
from intelliw.utils.intelliwapi import request
from intelliw.core.pipeline import Pipeline
from intelliw.utils.global_val import gl


class Infer:
    """
    infer entrypoint
    """
    def __init__(self, path, reporter_addr=None):
        self.pipeline = Pipeline(reporter_addr)
        self.pipeline.importmodel(path)

    def infer(self, data, request=request.Request(), func='infer', need_feature=True):
        """
        infer
        """
        return self.pipeline.infer(data, request, func, need_feature)




'''
Author: Hexu
Date: 2022-08-24 16:52:17
LastEditors: Hexu
LastEditTime: 2023-03-27 14:07:11
FilePath: /iw-algo-fx/intelliw/interface/trainjob.py
Description: Train entrypoint
'''
import traceback
from intelliw.core.al_decorator import flame_prof_process
from intelliw.core.trainer import Trainer
from intelliw.core.recorder import Recorder
from intelliw.datasets.datasets import get_dataset, DataSets, MultipleDataSets
import intelliw.utils.message as message
from intelliw.utils.logger import _get_framework_logger

logger = _get_framework_logger()


class TrainServer:
    def __init__(self, path, dataset_cfg, response_addr=None):
        self.response_addr = response_addr
        self.path = path
        self.reporter = Recorder(response_addr)
        self.dataset_cfg = dataset_cfg

    @flame_prof_process
    def run(self):
        try:
            trainer = Trainer(self.path, self.response_addr)
            datasets = get_dataset(self.dataset_cfg)
            # 检查是否支持流式, 1. DataSets 类型, 2 流式标志位, 3 数据源类型
            if isinstance(datasets, DataSets):
                logger.info("数据集类型: {}, 流式标志位: {}, 数据源类型: {}".format(type(datasets), datasets.is_streaming, datasets.source_type))  
            elif isinstance(datasets, MultipleDataSets):
                logger.info("数据集类型: {}, 流式标志位: {}".format(type(datasets), datasets.is_streaming))
            if isinstance(datasets, DataSets) and datasets.is_streaming and datasets.support_streaming():
                logger.info("使用流式训练模式")
                trainer.train_streaming(datasets)
            else:
                logger.info("使用传统训练模式")
                trainer.train(datasets)

        except (Exception, SystemExit) as e:
            stack_info = traceback.format_exc()
            logger.error("训练执行错误 {}, stack:\n{}".format(e, stack_info))
            self.reporter.report(
                message.CommonResponse(500, "train_fail", "训练执行错误 {}, stack:\n{}".format(e, stack_info)))



'''
Author: Hexu
Date: 2022-03-30 11:47:31
LastEditors: Hexu
LastEditTime: 2023-04-06 14:31:34
FilePath: /iw-algo-fx/intelliw/utils/exception.py
Description: 错误类定义
'''
####### error class #######


import os


class LimitConcurrencyError(Exception):
    pass


class ExceptionNoStack(Exception):
    def ignore_stack(self):
        return True


class PipelineException(Exception):
    pass


class ModelLoadException(Exception):
    def __init__(self, msg) -> None:
        self.msg = msg

    def __str__(self) -> str:
        return '''模型加载异常\n
报错信息: {}
可能原因:
1 内存分配过小, 模型加载时服务崩溃
2 模型文件是否存在,加载模型的路径是否正确
3 是否补全load函数,load函数是否存在问题
4 如果使用checkpoint,检查代码是否正确加载checkpoint
5 是否使用ONNX模型作为推理模型
'''.format(self.msg)


class UnverifiedRuntimeException(Exception):

    def __str__(self) -> str:
        return '''算法使用了未验证的运行环境：
    1. 请使用ONNX模型进行服务推理
    2. 请勿使用Torch， Paddle，TensorFlow等框架运行模型
        '''

    def ignore_stack(self):
        return True


class DatasetException(Exception):
    def ignore_stack(self):
        return True


class InferException(Exception):
    pass


class FeatureProcessException(Exception):
    def ignore_stack(self):
        return True


class DataSourceDownloadException(Exception):
    def ignore_stack(self):
        return True


class LinkServerException(Exception):
    pass


class HttpServerException(Exception):
    pass


class CheckpointException(Exception):
    def __str__(self) -> str:
        return '''checkpoint保存模型异常，发生错误的可能：
    1. save()方法在save_checkpoint()方法之前调用,请在训练结束后调用save()方法保存模型
    2. save_checkpoint()方法多次调用
        '''

    def ignore_stack(self):
        return True

class SnapshotException(Exception):
    def __str__(self) -> str:
        return '''snapshot保存模型异常，发生错误的可能：
    1. save()方法在save_snapshot()方法之前调用,请在训练结束后调用save()方法保存模型
    2. save_snapshot()方法参数version设置为非整数类型
        '''

    def ignore_stack(self):
        return True



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Time    : 2024/2/21 12:31
# @Author  : HEXU
# @File    : utils.py
# @Description :
import hashlib
from intelliw.config import config


def key_builder(func, args, kwargs):
    key, req = None, args[0].infer_request
    use_cache, cache_read, cache_write = True, True, True

    if not req.json and not req.query and not req.form:
        return key, not cache_read, not cache_write

    if isinstance(req.json, dict) and 'use_cache' in req.json:
        use_cache = req.json.pop('use_cache')

    if not use_cache:
        cache_read = False

    key = config.INFER_ID + args[2] + f"{req.json}{req.query}{req.form}"
    key = f"{hashlib.md5(key.encode()).hexdigest()[8:-8]}"
    return key, cache_read, cache_write


def skip_cache_func(result):
    return result[-1] is not None



#!/usr/bin/env python
# coding: utf-8
'''
Author: hexu
Date: 2021-10-14 14:54:05
LastEditTime: 2023-05-25 14:30:29
LastEditors: Hexu
Description: Algorithm -> DataSourceReader -> DataSource
FilePath: /iw-algo-fx/intelliw/datasets/datasource_base.py
'''

from collections.abc import Iterable
from abc import ABCMeta, abstractmethod
from operator import is_
from intelliw.config import config
from intelliw.utils.logger import _get_framework_logger


logger = _get_framework_logger()


class DataSourceType:
    '''输入数据源类型'''
    # 大类
    # 空
    EMPTY = 0  
    # 表格类别
    TABLE_TYPE = 6  
    TABLE_TYPE_LIST = [1, 2, 3, 9]
    # 图像类别
    CV_TYPE = 7  
    CV_TYPE_LIST = [4]
    # 文本类别
    NLP_TYPE = 8  
    NLP_TYPE_LIST = [21]
    # 大模型类别
    LARGE_MODEL_LIST = [24]

    # 子类
    REMOTE_CSV = 1  # 远程csv
    INTELLIV = 2  # 智能分析
    LOCAL_CSV = 3  # 本地 csv
    IW_IMAGE_DATA = 4  # 图片数据源
    IW_FACTORY_DATA = 5  # 数据工场数据集
    NLP_CORPORA = 21  # nlp语料
    SEMANTIC = 9  # 语义模型

    LARGE_MODEL = 24  # 大模型

    IW_FACTORY_DR_DATA = 25 # 数据工厂物化表数据源

    SUPPORT_STREAMING = [SEMANTIC, IW_FACTORY_DR_DATA]

    SQL = 998
    USER = 999


class AlgorithmsType:
    '''算法类型'''
    CLASSIFICATION = 3  # 分类算法
    TIME_SERIES = 9  # 时间序列


class DatasetType:
    TRAIN = 'train_set'
    VALID = 'validation_set'
    TEST = 'test_set'


class AbstractDataSource(metaclass=ABCMeta):
    """
    数据源定义
    """

    @abstractmethod
    def total(self) -> int:
        """
        获取数据源总数据条数
        :return: 数据源总数据条数
        """
        pass

    @abstractmethod
    def reader(self, page_size=100000, offset=0, limit=0, transform_function=None,
               dataset_type='train_set') -> Iterable:
        """
        获取一个读取该数据源的 iterator
        :param page_size: 读取分页大小
        :param offset:    开始读取的数据 index
        :param limit:     读取条数
        :param transform_function: 转换函数
        :return: 数据源 iterator
        """
        pass


class DataSourceReaderException(Exception):
    def ignore_stack(self):
        return True


class DataSourceWriterException(Exception):
    def ignore_stack(self):
        return True


class AbstractDataSourceWriter(metaclass=ABCMeta):
    def __init__(self) -> None:
        self.table_columns = None

    @abstractmethod
    def write(self, data, starttime):
        pass

    @abstractmethod
    def check_metadata(self, columns, **kwargs):
        pass
    
    @abstractmethod
    def create_table(self, table_name, columns, **kwargs):
        pass


class DatasetSelector:
    """
    DatasetSelector 
    """
    func = {}
    args = {}

    @classmethod
    def register_func(cls, key, value, args):
        cls.func[key] = value
        cls.args[key] = args

    @classmethod
    def parse(cls, cfg: dict) -> AbstractDataSource:
        stype = cfg["SOURCE_TYPE"]
        stream_flag = cfg.get("STREAM")
        is_streaming = False
        if stream_flag:
            if isinstance(stream_flag, bool):
                is_streaming = stream_flag
            elif isinstance(stream_flag, str):
                is_streaming = stream_flag.lower() == 'true'
            if isinstance(stream_flag, int):
                is_streaming = stream_flag > 0
            else:
                logger.warning("Stream flag must be bool or int or string")
        
        func = cls.func[stype]
        args = cls.args[stype]
        logger.info('\033[33mDatasetSelector: %s\033[0m', func.__name__)
        try:
            kwargs = cls._parse_args(args, cfg)
        except Exception:
            errmsg = f"Source Type: {stype}, Need Parameters: [{list(args.values())}]"
            raise DataSourceReaderException(errmsg)
        return func(**kwargs), stype, is_streaming

    @staticmethod
    def _parse_args(args: dict, cfg: dict) -> dict:
        kwargs = {}
        for param, key in args.items():
            if key == "INPUT_DATA_SOURCE_TRAIN_TYPE":
                value = cfg.get(key, getattr(config, key))
            elif isinstance(key, list):
                for _key in key:
                    value = cfg.get(_key)
                    if value:
                        break
            else:
                value = cfg[key]
            kwargs[param] = value
        return kwargs



'''
Author: Hexu
Date: 2022-11-22 09:48:10
LastEditors: Hexu
LastEditTime: 2023-02-22 13:37:41
FilePath: /iw-algo-fx/intelliw/core/trainer.py
Description: Train core
'''
# coding: utf-8
from intelliw.config import config
from intelliw.datasets.datasets import DataSets, MultipleDataSets
from intelliw.core.pipeline import Pipeline


class Trainer:
    def __init__(self, path, reporter=None):
        self.pipeline = Pipeline(reporter)
        self.pipeline.importmodel(path, True, config.CHECKPOINT_MODE, config.BASE_MODEL_MODE)

    def train(self, datasets: (DataSets, MultipleDataSets)):
        if not isinstance(datasets, DataSets) and not isinstance(datasets, MultipleDataSets):
            raise TypeError("datasets has a wrong type, required: DataSets, MultipleDataSets, actually: {}"
                            .format(type(datasets).__name__))

        return self.pipeline.train(datasets)

    def train_streaming(self, datasets: (DataSets, MultipleDataSets)):
        if not isinstance(datasets, DataSets) :
            raise TypeError("datasets has a wrong type, required: DataSets, actually: {}"
                            .format(type(datasets).__name__))
        
        return self.pipeline.train_streaming(datasets)




import os
import sys
import traceback
import argparse
import intelliw.utils.iuap_request as requests
from intelliw.datasets.datasets import get_dataset
from intelliw.core.infer import Infer
from intelliw.core.trainer import Trainer
from intelliw.core.custom import Custom
from intelliw.utils.logger import get_logger
from intelliw.utils.gen_model_cfg import generate_model_config_from_algorithm_config as __generate
from intelliw.config import config as iwconfig, dataset_config

report_addr = None
logger = get_logger(level='DEBUG')

#########################################################
#                请按照需求填写以下配置                     #
#########################################################
# (必须)设置算法文件位置（algorithms.py根目录）
PACKAGE_PATH = "./"

algorithm_yaml_path = os.path.join(PACKAGE_PATH, 'algorithm.yaml')
model_yaml_path = os.path.join(PACKAGE_PATH, 'model.yaml')
iwconfig.RUNNING_MODE = 'SCAFFOLDING'


# 训练测试(通过算法框架调用train)
def debug_train():
    '''(debug_train时必填)
    设置训练测试数据
    example:
        DATASET_ID = "asdasdasdasdas"   数据集id,获取途径 AI工作坊数据集管理中获取id, 不使用传入None
        TRAIN_CSV_PATH = "localcsvpath" or ["localcsvpath1", "localcsvpath2"]  本地训练csv, 如果写了DATASET_ID, 优先DATASET_ID
        DATASET_RATIO = [0.7, 0.2, 0.1] 训练/验证/测试 数据集比例, 值在0-1之间
        iwconfig.DATA_SPLIT_MODE = 0    数据集划分模式, 0 顺序划分,1 随机划分, 2 根据特征列进行划分(需要在algorithm.yaml的超参中增加targetCol,来指明特征列下标)

        TRAIN_NLP_CORPORA = "/home/xxx/xxx/"   本地训练nlp语料文件夹, 如果写了DATASET_ID, 优先DATASET_ID

        # cv标注类型: 0-自有 1-labelme 2-voc 3-coco
        # nlp语料文件类型: 20-txt 21-csv 22-json
        iwconfig.INPUT_DATA_SOURCE_TRAIN_TYPE = 1

    '''
    DATASET_ID = None

    # csv
    TRAIN_CSV_PATH = None

    # nlp
    TRAIN_NLP_CORPORA = None  # 需要是文件夹

    # cv标注类型: 0-自有 1-labelme 2-voc 3-coco
    # nlp语料文件类型: 20-txt 21-csv 22-json
    iwconfig.INPUT_DATA_SOURCE_TRAIN_TYPE = 22

    DATASET_RATIO = [0.7, 0.2, 0.1]
    iwconfig.DATA_SPLIT_MODE = 1  # 0-顺序划分 1-随机划分
    iwconfig.DATA_RANDOM_SEED = 1024 # 

    dataset_config.set_dataset_config(DATASET_ID, DATASET_RATIO,
                                      TRAIN_CSV_PATH, TRAIN_NLP_CORPORA)

    try:
        __generate(algorithm_yaml_path, model_yaml_path)
        trainer = Trainer(PACKAGE_PATH)
        train_data = get_dataset(iwconfig.DATASET_INFO)
        trainer.train(train_data)
    except Exception as e:
        stack_info = traceback.format_exc()
        logger.error(
            "fail to execute pipeline and stack:\n{}".format(str(stack_info)))


# 推理测试
def debug_infer():
    '''(debug_infer时必填)
    设置推理训练数据
    example: INFER_DATA = {"data":[1,2,3]}
    '''
    INFER_DATA = None
    try:
        import asyncio
        __generate(algorithm_yaml_path, model_yaml_path)
        infer = Infer(PACKAGE_PATH)
        loop = asyncio.get_event_loop()
        result = loop.run_until_complete(infer.infer(INFER_DATA))
        print(result)
    except Exception as e:
        stack_info = traceback.format_exc()
        logger.error(
            "fail to execute pipeline and stack:\n{}".format(str(stack_info)))


# 多api测试
def debug_router():
    '''(debug_router时必填)
    设置多api
    example:
        INFER_DATA = {"data":base64}
        ROUTER_FUNC = api_test
    '''
    INFER_DATA = None
    ROUTER_FUNC = None

    try:
        __generate(algorithm_yaml_path, model_yaml_path)
        infer = Infer(PACKAGE_PATH)
        instance = infer.pipeline.instance
        result = getattr(instance, ROUTER_FUNC)(INFER_DATA)
        print(result)
    except Exception as e:
        stack_info = traceback.format_exc()
        logger.error(
            "fail to execute pipeline and stack:\n{}".format(str(stack_info)))


# http服务测试
def debug_infer_http_server():
    # (debug_infer_http_server时必填)设置http服务测试
    SERVER_PORT = 8888

    try:
        __generate(algorithm_yaml_path, model_yaml_path)
        sys.argv = ["api.main", "-p", PACKAGE_PATH, "--port", str(SERVER_PORT)]
        from intelliw.interface.iwapi import iwmain
        iwmain.api_service.run()
    except Exception as e:
        stack_info = traceback.format_exc()
        logger.error(
            "fail to execute apiService and stack:\n{}".format(str(stack_info)))


# http接口测试
def debug_server_available():
    '''(debug_server_available时必填)
    http接口测试
    example：
        URL = "http://xxx.xxx.xxx/predict"
        METHOD = "get/post"
        INFER_DATA = {"data":base64}
    '''
    URL = "https://127.0.0.1:8888/predict"
    # METHOD = "post"
    INFER_DATA = None

    # resp = requests.request(METHOD, URL, json=INFER_DATA, timeout=60)
    resp = requests.post_json(url=URL, json=INFER_DATA, timeout=60)
    if resp.status != 200:
        resp.raise_for_status()
    print(resp.body)


def debug_custom():
    # 任务类型
    FRAMEWORK_MODE = "infer"  # importalg/importmodel/infer/train

    try:
        __generate(algorithm_yaml_path, model_yaml_path)
        custom = Custom(PACKAGE_PATH, procedure_type=FRAMEWORK_MODE)
        custom.run()
    except Exception as e:
        stack_info = traceback.format_exc()
        logger.error(
            "fail to execute pipeline and stack:\n{}".format(str(stack_info)))

func_dict = {
    'train': debug_train,
    'infer': debug_infer,
    'router': debug_router,
    'infer_http_server': debug_infer_http_server,
    'server_available': debug_server_available,
    'custem': debug_custom,
}

def run(args):
    job_type = args.job_type
    if job_type not in func_dict:
        logger.error("job type %s not found", job_type)
    else:
        func_dict[job_type]()


if __name__ == '__main__':
    if PACKAGE_PATH is None:
        raise FileNotFoundError("请在debug_controller.py中设置PACKAGE_PATH")

    parser = argparse.ArgumentParser(description=__doc__)
    parser.add_argument("-j", "--job_type", default="infer", type=str, help="job type")
    args = parser.parse_args()

    try:
        run(args)
        # debug_train()
        # debug_infer()
        # debug_router()
        # debug_infer_http_server()
        # debug_server_available()
        # debug_custom
    except (ImportError, TypeError) as e:
        print(
            "\033[1;31;40m如果为debug文件错误， 请通过命令 `intelliw init_debug` 更新debug文件\033[0m")
        traceback.print_exc()



import os
import time
import requests
import threading
from typing import Any, Optional, Dict
import zipfile
import shutil
import traceback

from .base import ModelLoader
from intelliw.utils.iuap_request import get as request_get
from intelliw.utils.iuap_request import get_modelfiles_by_modelid
from intelliw.utils.logger import _get_framework_logger
from intelliw.utils.storage_service import StorageService
logger = _get_framework_logger()

class APIModelLoader(ModelLoader):
    """
    从API加载模型的加载器
    """

    def __init__(self, config_manager: 'ConfigManager', load_func):
        """
        初始化API模型加载器

        :param config_manager: 配置管理器
        """
        self.config_manager = config_manager
        self._lock = threading.RLock()
        self._last_checked = 0
        self.load_func = load_func
        self.model_tmp_dir = "./model_tmp"
        os.makedirs(self.model_tmp_dir, exist_ok=True)

    def parse_latest_modelfile_info(self, resp_json):
        if resp_json is None:
            return None
        # todo
        resp_data = resp_json.get("data", {})
        resp_content = resp_data.get("content", [])
        if len(resp_content) > 0:
            modelfile_info = {}
            modelfile_info['version'] = resp_content[0].get("instanceVersion")
            modelfile_info['path'] = resp_content[0].get("modelPath")
            modelfile_info['lastModifiedDate'] = resp_content[0].get("lastModifiedDate", 0)
            return modelfile_info
        return None

    def load(self, model_path: str, version: str) -> Any:
        """
        从API加载模型

        :param model_path: 模型路径
        :return: 加载的模型
        """
        try:

            logger.info(f"Download model from {model_path}")
            local_download_path=f'{self.model_tmp_dir}/tmp_model_v_{version}_{int(time.time() * 1e6)}'
            downloader = StorageService(
                model_path, "download")
            downloader.download(local_download_path, stream=True)
            
            if not os.path.exists(local_download_path):
                raise Exception(f"Failed to download model from {model_path}, version: {version}")

            logger.info(f"Successfully loaded model: {model_path}, version: {version}")
            #  加载模型

            target_model_path = local_download_path
            if zipfile.is_zipfile(local_download_path):
                # 创建临时目录来解压ZIP文件
                extract_dir = local_download_path + "_extracted"
                os.makedirs(extract_dir, exist_ok=True)
                
                # 解压ZIP文件到临时目录
                with zipfile.ZipFile(local_download_path) as zip_file:
                    zip_file.extractall(extract_dir)
                
                target_model_path = extract_dir
            
            model = self.load_model(target_model_path)
            if not model:
                logger.error(f"Failed to load model: {model_path}, version: {version}")
            logger.info(f"Successfully validated model: {model_path}, version: {version}")
            try:
                # 如果存在解压目录，删除它  
                if 'extract_dir' in locals() and os.path.exists(extract_dir):
                    shutil.rmtree(extract_dir)
                    logger.info(f"Deleted temporary directory: {extract_dir}")
                
                # 删除下载的原始文件
                if os.path.exists(local_download_path):
                    if os.path.isfile(local_download_path):
                        os.remove(local_download_path)
                    elif os.path.isdir(local_download_path):
                        shutil.rmtree(local_download_path)
                    logger.info(f"Deleted temporary file: {local_download_path}")
            except Exception as e:
                logger.warning(f"Failed to delete temporary files: {str(e)}")
            return model

        except Exception as e:
            logger.error(f"Error loading model: {str(e)}")
            raise
    
    def load_model(self, model_path: str) -> Any:
        """
        加载模型

        :param model_path: 模型路径
        :return: 加载的模型
        """
        try:
            model = self.load_func(model_path)
            return model
        except Exception as e:
            logger.error(f"Error loading model: {str(e)}")
            return None

    def validate(self, model: Any) -> bool:
        """
        验证模型有效性

        :param model: 模型对象
        :return: 是否有效
        """
        if not model:
            return False

        return True

    def check_for_updates(self) -> Optional[Dict[str, str]]:
        """
        检查模型更新

        :return: 包含新模型路径和版本的字典，如果没有更新则返回None
        """
        try:
            # 检查时间间隔
            update_interval = self.config_manager.get('MODEL_UPDATE_INTERVAL', 300)
            current_time = time.time()

            # 如果距离上次检查时间不足更新间隔，则跳过
            with self._lock:
                if current_time - self._last_checked < update_interval:
                    return None
                self._last_checked = current_time

            
            # 请求接口
            url = self.config_manager.get('MODELFILES_BY_MODELID_URL')
            model_id = self.config_manager.get('TRAININGASSIGNMENT_MODEL_ID')
            # 设置超时
            timeout = self.config_manager.get('MODEL_UPDATE_TIMEOUT', 30)
            resp_json = get_modelfiles_by_modelid(url, model_id, timeout)
            logger.info(f"get_modelfiles_by_modelid response: {resp_json}")
            # 解析接口， 获取最新的模型文件信息
            latest_modelfile_info = self.parse_latest_modelfile_info(resp_json)
            if not latest_modelfile_info:
                logger.info(f"No models found in update response, response: {resp_json}")
                return None
            
            logger.info(f"latest_modelfile_info: {latest_modelfile_info}")

            current_version = self.config_manager.get('CURRENT_MODEL_VERSION')
            current_modified_date = self.config_manager.get('CURRENT_MODEL_MODIFIED_DATE')

            # 如果没有当前版本或者版本不同, 版本相同但修改时间更新，则认为有更新
            if not current_version \
                or latest_modelfile_info['version'] != current_version \
                or latest_modelfile_info['lastModifiedDate'] > current_modified_date:
                logger.info(f"Found new model version: {latest_modelfile_info['version']}, new modified date: {latest_modelfile_info['lastModifiedDate']}")
                return {
                    'path': latest_modelfile_info['path'],
                    'version': latest_modelfile_info['version'],
                    "modified_date": latest_modelfile_info['lastModifiedDate'],
                }

            logger.info("No new model updates available")
            return None

        except Exception as e:
            traceback.print_exc()
            logger.error(f"Error checking for model updates: {e}")
            return None


from dataclasses import dataclass, asdict


@dataclass
class DatasetConfigParams:
    """
    数据集配置
    """
    sourceType: int = 0  # 1-csv 9-语义模型
    dataSourceAddress: str = None  # csv下载地址
    inputAddr: str = None  # 获取数据接口
    inputGetrowAddr: str = None  # 获取数据接口
    inputDataMetaAddr: str = None  # 获取字段名接口
    inputModelId: str = None  # 语义模型id
    tenantId: str = None  # 租户id
    tableId: str = None  # 表id
    tableName: str = None  # 表名
    inputDataSourceId: str = None  # 数据工场表id
    userId: str = None  # 用户id
    stream: bool = False  # 流式标志位

@dataclass
class DatasetConfig:
     '''
     数据集配置
     '''
     SOURCE_TYPE: int = 0
     DATA_SOURCE_ADDRESS: str = None  # csv下载地址
     INPUT_ADDR: str = None  # 获取数据接口
     INPUT_GETROW_ADDR: str = None  # 获取数据接口
     INPUT_DATA_META_ADDR: str = None  # 获取字段名接口
     INPUT_MODEL_ID: str = None  # 语义模型id
     TENANT_ID: str = None  # 租户id
     TABLE_ID: str = None  # 表id
     TABLE_NAME: str = None  # 表名
     INPUT_DATA_SOURCE_ID: str = None  # 数据工场表id
     USER_ID: str = None  # 用户id
     STREAM: bool = False  # 流式标志位


def convert(dataset_config_params: DatasetConfigParams) -> DatasetConfig:
    """
    转换数据集配置参数为数据集配置对象
    驼峰格式转下划线格式
    """
    return DatasetConfig(
        SOURCE_TYPE=dataset_config_params.sourceType,
        DATA_SOURCE_ADDRESS=dataset_config_params.dataSourceAddress,
        INPUT_ADDR=dataset_config_params.inputAddr,
        INPUT_GETROW_ADDR=dataset_config_params.inputGetrowAddr,
        INPUT_DATA_META_ADDR=dataset_config_params.inputDataMetaAddr,
        INPUT_MODEL_ID=dataset_config_params.inputModelId,
        TENANT_ID=dataset_config_params.tenantId,
        TABLE_ID=dataset_config_params.tableId,
        TABLE_NAME=dataset_config_params.tableName,
        INPUT_DATA_SOURCE_ID=dataset_config_params.inputDataSourceId,
        USER_ID=dataset_config_params.userId,
        STREAM=dataset_config_params.stream
    )
    
    



import uuid
from typing import Dict, Optional, List
from concurrent.futures import ThreadPoolExecutor
import threading
import atexit
import time
import sys

    # 定义必填字段列表
REQUIRED_FIELDS = ['traceId', 'apiCode', 'resultCode', 'startTime', 'ytenantId', 'domain']

    # 定义字段格式校验规则
VALID_FORMAT_FIELDS = {
        "requestHeader": Optional,  # 可为 None 或字符串
        "requestBody": Optional,    # 可为 None 或字符串
        "responseBody": Optional,   # 可为 None 或字符串
    }

class TraceLog:
    _executor = None
    _lock = threading.Lock()
    
    @classmethod
    def _get_executor(cls):
        if cls._executor is None:
            with cls._lock:
                if cls._executor is None:
                    cls._executor = ThreadPoolExecutor(
                        max_workers=2,
                        thread_name_prefix="TraceLog_"
                        )
                    atexit.register(cls._shutdown_executor)
        return cls._executor

    

    @classmethod
    def _shutdown_executor(cls):
        """安全关闭线程池"""
        if cls._executor:
            cls._executor.shutdown(wait=True)
            cls._executor = None
    def __init__(self):
        # 基础跟踪信息
        self.id: Optional[str] = str(uuid.uuid4())
        self.traceId: Optional[str] = None  # 必填
        self.spanId: Optional[str] = None
        self.pSpanId: Optional[str] = None  # 必填
        
        # 请求/响应信息
        self.queryMsg: Optional[str] = None
        self.apiCode: Optional[str] = None  # 必填
        self.apiUrl: Optional[str] = None
        self.resultCode: Optional[int] = None  # 必填
        self.requestHeader: Optional[str] = None
        self.requestBody: Optional[str] = None
        self.responseBody: Optional[str] = None
        
        # 时间信息
        self.startTime: Optional[int] = int(time.time() * 1000)  # 必填，时间戳（毫秒）
        self.endTime: Optional[int] = None    # 时间戳（毫秒）
        self.timeCost: Optional[int] = None   # 耗时（毫秒）
        
        # 监控与业务信息
        self.entranceFlag: Optional[int] = 0  # 入口标志 (0/1)
        self.warnThreshold: Optional[int] = None  # 告警阈值（毫秒）
        self.domain: Optional[str] = 'iuap-aip-alg'  # 必填
        self.businessKey: Optional[str] = None
        self.businessType: Optional[str] = None
        self.businessData: Dict[str, str] = {}  # 并发安全的字典
        
        # 租户与创建信息
        self.ytenantId: Optional[str] = None  # 必填
        self.creatorId: Optional[str] = None
        self.creator: Optional[str] = None
        self.createTime: Optional[int] = int(time.time() * 1000)   # 创建时间戳（毫秒）
        self.ts: Optional[int] = None          # 时间戳（毫秒）

    def validate_required_fields(self) -> None:
        """验证必填字段是否已赋值"""
        missing_fields = []
        for field in REQUIRED_FIELDS:
            value = getattr(self, field)
            # 检查字段是否为None或空字符串
            if value is None or (isinstance(value, str) and value.strip() == ""):
                missing_fields.append(field)
        
        if missing_fields:
            raise ValueError(f"缺少必填字段: {', '.join(missing_fields)}")
    def validate_field_formats(self) -> None:
        """验证字段格式是否符合要求"""
        format_errors = []
        
        for field, expected_format in VALID_FORMAT_FIELDS.items():
            value = getattr(self, field)
            
            # 检查字段格式
            if expected_format is Optional:
                # 检查是否为字符串或None
                if value is not None and not isinstance(value, str):
                    format_errors.append(
                        f"字段 '{field}' 格式错误: 应为字符串或None, 实际类型 {type(value).__name__}"
                    )
        
        if format_errors:
            raise TypeError("\n".join(format_errors))
        
    def full_validation(self) -> None:
        """执行完整的字段验证（必填字段 + 格式校验）"""
        self.validate_required_fields()
        self.validate_field_formats()
        self.validate_val_length()

    def validate_val_length(self):
        if self.requestBody and sys.getsizeof(self.requestBody) >= 1024 * 1024 * 1:
            self.requestBody = "RequestBody too long, Not Logged!!!"

        if self.responseBody and sys.getsizeof(self.responseBody) >= 1024 * 1024 * 1:
            self.responseBody  = "ResponseBody too long, Not Logged!!!"

    def calculate_time_cost(self) -> None:
        """自动计算请求耗时（需先设置 startTime 和 endTime）"""
        if self.startTime and self.endTime:
            self.timeCost = self.endTime - self.startTime

    def add_business_data(self, key: str, value: str) -> None:
        """线程安全地添加业务扩展数据"""
        self.businessData[key] = value

    def to_dict(self) -> Dict:
        """将对象转换为字典（包含必填字段验证）"""
        self.full_validation()
        return {attr: getattr(self, attr) for attr in dir(self) 
                if not callable(getattr(self, attr)) and not attr.startswith("_")}



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Time    : 2024/2/22 16:38
# @Author  : HEXU
# @File    : common.py
# @Description :


mode_set = {'single', 'sentinel', 'cluster'}


def check_mode(mode: str):
    return mode.lower() in mode_set



import datetime
from typing import Iterable
from intelliw.datasets.datasource_base import AbstractDataSource, DatasetSelector, \
    DataSourceType, AbstractDataSourceWriter


class EmptyDataSourceWriter(AbstractDataSourceWriter):

    def write(self, data, starttime=None):
        if starttime is None:
            starttime = datetime.datetime.now()
        print(
            "datasource output: {}, starttime: {}".format(data, starttime))
        return {'status': 1}
    
    def check_metadata(self, columns, **kwargs):
        pass
    
    def create_table(self, table_name, columns, **kwargs):
        pass


class EmptyDataSource(AbstractDataSource):
    """
    空数据源
    """

    def __init__(self):
        pass

    def total(self):
        return 0

    def reader(self, page_size=100000, offset=0, limit=0, transform_function=None,
               dataset_type='train_set') -> Iterable:
        return self.__Reader()

    class __Reader:
        def __iter__(self):
            return self

        def __next__(self):
            raise StopIteration
        
        def get_meta(self):
            """获取元数据"""
            return []


DatasetSelector.register_func(DataSourceType.EMPTY, EmptyDataSource, {})



import os
import json
import yaml
from typing import Dict, Any, Optional

from intelliw.utils.logger import _get_framework_logger


from .base import ConfigManager
logger = _get_framework_logger()


class ModelManagerConfig(ConfigManager):
    def __init__(self):
        # 初始化配置字典
        modelfiles_by_modelid_url = f'{os.environ.get("domain.url")}/{os.getenv("YMS_CONSOLE_APP_CODE")}/self/rest/api/trainingInstanceModelFiles'
        self._config = {
            'MODELFILES_BY_MODELID_URL': os.environ.get('MODELFILES_BY_MODELID_URL', modelfiles_by_modelid_url),
            'MODEL_UPDATE_INTERVAL': int(os.environ.get('MODEL_UPDATE_INTERVAL', '300')),  # 默认5分钟
            'REQUEST_TIMEOUT': int(os.environ.get('REQUEST_TIMEOUT', '10')),  # 默认10秒
            'TRAININGASSIGNMENT_MODEL_ID': os.environ.get('TRAININGASSIGNMENT_MODEL_ID'),
            'MODEL_LOAD_TIMEOUT': int(os.environ.get('MODEL_LOAD_TIMEOUT', '30')),
            'MODEL_UPDATE_TIMEOUT': int(os.environ.get('MODEL_UPDATE_TIMEOUT', '30')),
            'CURRENT_MODEL_VERSION': None,
            'CURRENT_MODEL_MODIFIED_DATE': None,
            'HOT_RELOAD_MODE': os.environ.get('HOT_RELOAD_MODE', 'false').lower() == 'true',
        }

    def load_config(self, config_path: Optional[str] = None) -> Dict[str, Any]:
        """
        加载配置

        :param config_path: 配置文件路径
        :return: 配置字典
        """
        if config_path and os.path.exists(config_path):
            try:
                with open(config_path, 'r', encoding='utf-8') as f:
                    if config_path.endswith('.json'):
                        file_config = json.load(f)
                    elif config_path.endswith('.yaml') or config_path.endswith('.yml'):
                        file_config = yaml.safe_load(f)
                    else:
                        raise ValueError(f"Unsupported config file format: {config_path}")

                # 更新配置
                for key, value in file_config.items():
                    self._config[key] = value

                logger.info(f"Config loaded from {config_path}")
            except Exception as e:
                logger.error(f"Error loading config from {config_path}: {str(e)}")

        # 从环境变量更新配置
        for key in self._config:
            if key in os.environ:
                # 尝试转换类型
                try:
                    current_value = self._config[key]
                    logger.info(f"Updating config {key} from environment variable, old value: {current_value}")
                    if isinstance(current_value, int):
                        self._config[key] = int(os.environ[key])
                    elif isinstance(current_value, float):
                        self._config[key] = float(os.environ[key])
                    elif isinstance(current_value, bool):
                        self._config[key] = os.environ[key].lower() == 'true'
                    else:
                        self._config[key] = os.environ[key]
                except ValueError as e:
                    logger.warning(f"Failed to convert environment variable {key} to appropriate type: {str(e)}")
                    self._config[key] = os.environ[key]

        return self._config.copy()

    def get(self, key: str, default: Any = None) -> Any:
        """
        获取配置项

        :param key: 配置键
        :param default: 默认值
        :return: 配置值
        """
        return self._config.get(key, default)

    def set(self, key: str, value: Any) -> None:
        """
        设置配置项

        :param key: 配置键
        :param value: 配置值
        """
        self._config[key] = value

    def to_dict(self) -> Dict[str, Any]:
        """
        转换为字典

        :return: 配置字典
        """
        return self._config.copy()


# # 创建配置实例
# config = ModelManagerConfig()
# # 加载配置
# config.load_config()


"""
Author: Hexu
Date: 2022-05-05 12:07:16
LastEditors: Hexu
LastEditTime: 2024/12/2 14:50
FileName: __init__
Description: 
"""
from ._context import ContextThreadExecutor, ContextThread, header_ctx

__all__ = ["ContextThreadExecutor", "ContextThread", "header_ctx"]



import logging
import time
from copy import deepcopy

from intelliw.utils.intelliwapi import _data_pool_context_storage, _request_header_context_storage, \
    _request_pool_context_storage
from contextvars import copy_context, ContextVar
import threading
from concurrent.futures import ThreadPoolExecutor, Future
from functools import partial


class _Context:
    def __init__(self, instance: ContextVar):
        self.ctx = instance

    def get(self):
        try:
            return self.ctx.get()
        except LookupError as e:
            raise RuntimeError(
                "You didn't use ContextMiddleware or "
                "you're trying to access `context` object "
                "outside of the request-response cycle."
            ) from e

    def copy_context(self):
        return copy_context()

    def copy_set(self, data):
        return self.ctx.set(
            deepcopy(data)
        )

    def set(self, data):
        return self.ctx.set(
            data
        )

    def exists(self) -> bool:
        return self.ctx in copy_context()


# context = _Context(_data_pool_context_storage)
header_ctx = _Context(_request_header_context_storage)
req_ctx = _Context(_request_pool_context_storage)


class CurrentReq:
    def __init__(self, request=None):
        if req_ctx:
            req_ctx.set(request)

    def __getattr__(self, key):
        return getattr(req_ctx.get(), key)

    def __setattr__(self, key, value):
        req = req_ctx.get()
        setattr(req, key, value)
        req_ctx.set(req)


class ContextThread(threading.Thread):
    def __init__(self, group=None, target=None, name=None,
                 args=(), kwargs=None, *, daemon=None):
        ctx = copy_context()
        # 包装目标函数
        super().__init__(target=ctx.run, args=(target, *args), kwargs=kwargs, group=group, name=name, daemon=daemon)


class ContextThreadExecutor:
    def __init__(self, max_workers=None, thread_name_prefix='',
                 initializer=None, initargs=()):
        self.executor = ThreadPoolExecutor(max_workers=max_workers, thread_name_prefix=thread_name_prefix,
                                           initializer=initializer, initargs=initargs)

    def submit(self, func, *args, **kwargs) -> Future:
        ctx = copy_context()
        return self.executor.submit(ctx.run, partial(func, *args, **kwargs))

    def map(self, fn, *iterables, timeout=None, chunksize=1):
        if timeout is not None:
            end_time = timeout + time.monotonic()

        fs = [self.submit(fn, *args) for args in zip(*iterables)]

        def result_iterator():
            try:
                fs.reverse()
                while fs:
                    if timeout is None:
                        yield fs.pop().result()
                    else:
                        yield fs.pop().result(end_time - time.monotonic())
            finally:
                for future in fs:
                    future.cancel()

        return result_iterator()

    def shutdown(self, wait=True, *args, **kwargs):
        return self.executor.shutdown(wait=wait, *args, **kwargs)

    def __getattr__(self, name):
        return getattr(self.executor, name)

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.shutdown()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Time    : 2024/2/20 19:35
# @Author  : HEXU
# @File    : connect.py
# @Description :
import os

import redis.retry
from time import sleep
from redis.client import Redis
from redis.sentinel import Sentinel
from redis.cluster import RedisCluster, ClusterNode
from redis.exceptions import ConnectionError, TimeoutError
from intelliw.utils.logger import _get_framework_logger
from intelliw.utils.redis import common
from intelliw.utils.util import use_ipv6_by_env

logger = _get_framework_logger()


class RedisConnect:
    def __init__(self, client_type, url, password, sentinel_password=None, master_name=None, db=0, pool_size=10,
                 retry_attempts=3, retry_delay=2,host=None, port = None):
        try:
            self.db = db
            self.url = url
            self.redis_client = None
            self.type = client_type
            self.password = password
            self.pool_size = int(pool_size) if pool_size else 10
            self.sentinel_password = sentinel_password
            self.retry_attempts = retry_attempts
            self.retry_delay = retry_delay
            self.master_name = master_name
            self.host = host
            self.port = port

            if not common.check_mode(self.type):
                logger.error("mode is not supported")
                return

            if self.type == "single":
                if not self.host and not self.port:
                    self.host, self.port = self.get_host(self.url)
                self.redis_client = Redis(host=self.host, port=self.port, db=self.db,
                                          password=self.password, decode_responses=False,
                                          max_connections=self.pool_size, retry_on_timeout=True)

            elif self.type == "cluster":
                self.nodes = []
                for url in self.url.split(','):
                    host, port = self.get_host(url)
                    self.nodes.append(ClusterNode(host=host, port=port))

                self.redis_client = RedisCluster(startup_nodes=self.nodes, decode_responses=False,
                                                 password=self.password, max_connections=self.pool_size,
                                                 socket_timeout=2, retry_on_timeout=True)

            elif self.type == "sentinel":
                self.nodes = [self.get_host(url) for url in self.url.split(',')]
                self.redis_sentinel = Sentinel(self.nodes, sentinel_kwargs={'password': self.sentinel_password},
                                               socket_timeout=1.0)
                self.redis_client = self._get_sentinel_client()

            self._ping_with_retries()

        except Exception as e:
            self.redis_client = None
            logger.error(f"Error initializing Redis client: {e}")

    def get_host(self, url):
        # if self.is_ipv6:
        split = url.rindex(":")
        port = url[split + 1:]
        host = url[:split].replace("]", "").replace("[", "")
        # else:
            # host, port = self.url.split(':')
        return host, int(port)

    def _get_sentinel_client(self):
        try:
            return self.redis_sentinel.master_for(self.master_name, socket_timeout=1.0,
                                                  password=self.password, db=self.db,
                                                  decode_responses=False, max_connections=self.pool_size)
        except ConnectionError as e:
            logger.error(f"Error connecting to Redis Sentinel master: {e}")
            return None

    def _ping_with_retries(self):
        """重试机制，在 Redis 启动时进行 ping 以确保连接可用"""
        attempts = 0
        while attempts < self.retry_attempts:
            try:
                if self.redis_client:
                    self.redis_client.ping()
                    logger.info("Redis connection established.")
                    break
            except (ConnectionError, TimeoutError) as e:
                attempts += 1
                logger.warning(f"Attempt {attempts} failed: {e}. Retrying in {self.retry_delay} seconds...")
                sleep(self.retry_delay)
        else:
            logger.error("Failed to connect to Redis after multiple attempts.")
            self.redis_client = None

    def __del__(self):
        pass



from contextvars import ContextVar
from intelliw.utils.intelliwapi.request import Request

_data_pool_context_storage: ContextVar = ContextVar(
    "data-pool", default=None
)

_request_header_context_storage: ContextVar = ContextVar(
    "request-header", default={}
)

_request_pool_context_storage: ContextVar = ContextVar(
    "req-pool", default=Request()
)



import re

MATCH_STR = r'\$\{([\w._-]+)\}'

def parse_replace_config(raw_dict):
    # 初始化两个字典
    todo_dict = raw_dict.copy()
    done_dict = {}
    max_iterations = 10  # 最大循环次数
    iteration = 0
    progress = True  # 标记每次循环是否有所进展

    while todo_dict and iteration < max_iterations and progress:
        progress = False
        keys_to_process = list(todo_dict.keys())  # 遍历前获取当前所有键的副本
        # print("="*20)
        # print(f"循环次数 {iteration}，剩余待处理键：{todo_dict.keys()}")
        for key in keys_to_process:
            current_value = todo_dict[key]
            # 尝试替换所有变量引用（使用 done_dict 中的已知值）
            new_value = replace_variables(current_value, done_dict)
            
            # 如果替换后不再包含变量
            if not has_variables(new_value):
                # 将处理好的键移到 done_dict
                done_dict[key] = new_value
                del todo_dict[key]
                progress = True  # 标记有进展
            elif new_value != current_value:
                todo_dict[key] = new_value
                progress = True
            else:
                pass
                # print(f"循环次数 {iteration}，无法替换变量 {key} 的值：{current_value}")
                

        iteration += 1
    # 
    if todo_dict:
        for key, value in todo_dict.items():
            done_dict[key] = value

    return done_dict

def replace_variables(value, done_dict):
    """递归替换变量，直到无法再替换"""
    while True:
        matches = re.findall(MATCH_STR, value)
        if not matches:
            break
        replaced = False
        for var in matches:
            # print(f"value: {value}, match: {var}")
            if var in done_dict:
                replacement = done_dict[var]
                value = value.replace(f"${{{var}}}", replacement)
                replaced = True
        if not replaced:
            break  # 没有可替换的变量，退出循环
    return value

def has_variables(value):
    """检查值中是否还存在未替换的变量"""
    return bool(re.search(MATCH_STR, value))

if __name__ == "__main__":
    # 示例数据
    data = [
        "udhsdweb-url=https://${udhsdweb-host}",
        "udhsdweb-host=yssdweb-bip-test.yonyoucloud.com",
        "a=https://${udhsdweb-host}",
        "public.domain.gpaas.url=${public.domain.schema}${public.domain.gpaas.host}",
        "public.domain.url=${public.domain.schema}${public.domain.host}",
        "ymc.public.domain.url=${public.domain.gpaas.url}",
        "ymc.public.domain.url1=${ymc.public.domain.url1}",
        "public.domain.gpaas.host=yms-foreingn-gpaas-w.yyuap.com:9011",
        "public.domain.host=bip-test-xxx.yyuap.com",
        "public.domain.schema=https://",
        "domain.yonbip-scm-scmsa_1=https://bip-test.yonyoucloud.com/yonbip-scm-scmsa",
        "support.udh.udhOpenUrl=${domain.yonbip-scm-scmsa_1}",

    ]
    data = {key: value for value in data for key, value in [line.split('=', 1) for line in data]}

    # 调用函数并输出结果
    result = parse_replace_config(data)
    print(result)


'''
Author: Hexu
Date: 2022-09-26 13:34:19
LastEditors: Hexu
LastEditTime: 2022-09-27 10:01:01
FilePath: /iw-algo-fx/intelliw/utils/database/__init__.py
Description: 
'''
try:
    import sqlalchemy
except ImportError:
    raise ImportError("\033[31mIf use database, you need: pip install sqlalchemy (>=1.4) \033[0m")

from .connection import get_engine
from .connection import get_scoped_session
from .schema import SchemaUtils
from .proxy import make_session_proxy



#!/usr/bin/env python
# coding: utf-8
import asyncio
import json
import time
import traceback

from fastapi import FastAPI, Request as FARequest, Response as FAResponse, BackgroundTasks
from intelliw.utils.crontab import Crontab
from intelliw.datasets.datasets import get_dataset, get_datasource_writer, \
    DataSets, MultipleDataSets
from intelliw.core.recorder import Recorder
from intelliw.core.infer import Infer
from intelliw.utils import message
import uvicorn
from intelliw.datasets.datasource_base import DataSourceReaderException, DataSourceWriterException
from intelliw.utils import get_json_encoder, intelliwapi
from intelliw.config import config
from intelliw.utils.logger import _get_framework_logger

logger = _get_framework_logger()

batchjob_infer = "batchjob-infer"

INFER_CONTROLLER: Infer = None  # type: ignore
intelliw_setting = {}


def get_batch_msg(issuccess, msg, status=True, starttime=None, user_param={}, system_param={}):
    """
    批处理上报信息，inferTaskStatus 不为空，会被记录为一次调用，标识一次批处理的状态
    """
    infer_task_status = []
    if status:
        infer_task_status = [
            {
                "id": config.INFER_ID, "issuccess": issuccess,
                "starttime": starttime, "endtime": int(time.time() * 1000),
                "message": msg
            }
        ]
    out_msg = [
        {
            'status': 'start' if issuccess else 'end',
            'inferid': config.INFER_ID,
            'instanceid': config.INSTANCE_ID,
            'inferTaskStatus': infer_task_status,
            "params": {"user": user_param, "system": system_param}
        }
    ]
    return json.dumps(out_msg, cls=get_json_encoder(), ensure_ascii=False)


def validate_batch_job(reporter, path):
    global INFER_CONTROLLER
    INFER_CONTROLLER = Infer(path, reporter)
    msg = get_batch_msg(True, '定时推理校验通过，上线成功', status=False)
    reporter.report(message.CommonResponse(200, batchjob_infer,
                                           '定时推理校验通过，上线成功',
                                           json.dumps(msg, cls=get_json_encoder(), ensure_ascii=False)))


def infer_job(reporter, dataset_cfg, output_dataset_cfg, params={}):
    """
    request json {"user":None, "system": None}
    response json {
                    'status': 'start' if issuccess else 'end',
                    'inferid': config.INFER_ID,
                    'instanceid': config.INSTANCE_ID,
                    'inferTaskStatus': inferTaskStatus,
                    "params": {"user":None, "system": None}
                }
    """
    user_param = params.get("user", {})
    system_param = params.get("system", {})

    def input():
        datasets = get_dataset(dataset_cfg)
        alldata = datasets.read_all_data()
        # 保持与训练一致
        if isinstance(datasets, MultipleDataSets):
            return alldata
        elif isinstance(datasets, DataSets):
            return [alldata]
        else:
            return None

    def output(r):
        writer = get_datasource_writer(output_dataset_cfg)
        res_data = writer.write(r)
        if res_data['status'] != 1:
            raise DataSourceWriterException(res_data['msg'])

    start_time = int(time.time() * 1000)
    result, stack_info, info = None, None, ""
    try:
        # 输入
        alldata = input()

        # 批处理
        global INFER_CONTROLLER
        pipeline = INFER_CONTROLLER.pipeline
        infer_func = getattr(pipeline.instance, "infer")
        arg = [alldata] if infer_func.__code__.co_argcount > 1 else []  # 获取函数参数数量
        pipeline.instance.request = intelliwapi.Request(user_param=user_param)
        result = infer_func(*arg)
        logger.info('批处理处理结果 %s', result)

        # 输出
        output(result)
    except DataSourceReaderException as e:
        info = f"批处理输入数据错误: {e}"
        stack_info = f"{info}, stack:\n{traceback.format_exc()}"
    except DataSourceWriterException as e:
        info = f"批处理输出数据错误: {e}"
        stack_info = f"{info}, stack:\n{traceback.format_exc()}"
    except Exception as e:
        info = f"批处理执行错误: {e}"
        stack_info = f"{info}, stack:\n{traceback.format_exc()}"
    else:
        info = "批处理输出数据成功"
    finally:
        status_code = 200
        if stack_info is not None:
            status_code = 500
            logger.error(stack_info)

        msg = get_batch_msg(True, info, user_param=result,
                            system_param=system_param, starttime=start_time)
        reporter.report(
            message.CommonResponse(status_code, batchjob_infer, info, msg)
        )


class JobServer:
    @staticmethod
    async def healthcheck():
        resp = message.HealthCheckResponse(200, "api", 'ok', "")
        return FAResponse(content=str(resp), media_type="application/json")

    @staticmethod
    async def run(req: FARequest, background_tasks: BackgroundTasks):
        r = await req.json()
        args = intelliw_setting[f'api-batch-predict']
        logger.info("server job start")

        background_tasks.add_task(
            infer_job,
            args["reporter"],
            args["dataset_cfg"],
            args["output_dataset_cfg"],
            r
        )
        return FAResponse(content=json.dumps({"code": "1", "message": "server job start"}),
                          media_type="application/json")


class BatchService:
    def __init__(self, corn, path, dataset_cfg, output_dataset_cfg, response_addr=None, task='infer'):
        self.reporter = Recorder(response_addr)
        if task != 'infer':
            _msg = '批处理任务任务错误，TASK环境变量必须为infer'
            msg = get_batch_msg(False, _msg, status=False)
            self.reporter.report(message.CommonResponse(500, batchjob_infer,
                                                        _msg,
                                                        json.dumps(msg, cls=get_json_encoder(), ensure_ascii=False)))
        self.corn = corn
        self.only_server = not corn
        self.dataset_cfg = dataset_cfg
        self.output_dataset_cfg = output_dataset_cfg
        self.path = path
        intelliw_setting['api-batch-predict'] = {"reporter": self.reporter, "dataset_cfg": self.dataset_cfg,
                                                 "output_dataset_cfg": self.output_dataset_cfg}

    def _format_parse(self):
        return [{
            'crontab': f.strip(),
            'func': infer_job,
            'args': (self.reporter, self.dataset_cfg, self.output_dataset_cfg)
        } for f in self.corn.split("|")]

    def _cronjob(self):
        job_list = self._format_parse()
        crontab = Crontab(job_list, True)
        crontab.start()
        logger.info("Start cronjob")

    def _server(self):
        app = FastAPI(
            json_encoder=get_json_encoder()
        )
        app.add_api_route(
            '/batch-predict', endpoint=JobServer.run, methods=['get', 'post'])
        app.add_api_route(
            "/healthcheck", endpoint=JobServer.healthcheck, methods=["POST", "GET"])
        logger.info(
            "Server Start: \n\033[33m[POST] /batch-predict\n[POST, GET] /healthcheck\033[0m"
        )
        asyncio.set_event_loop(asyncio.new_event_loop())
        uvicorn.run(
            app=app,
            host="0.0.0.0",
            port=8888,
        )

    def run(self):
        validate_batch_job(self.reporter, self.path)
        if self.only_server:
            logger.info(
                "\033[33mCronjob Format is Empty, Only Server Mode\033[0m")
            self._server()
        else:
            logger.info(
                "\033[33mCronjob Format is %s,Cronjob and Server Mode\033[0m", self.corn)
            self._cronjob()
            self._server()



import os
import time
import threading
from typing import Any, Optional, Dict

# 导入基础类和单例装饰器
from .base import BaseModelManager
from .singleton import singleton
from .model_loader import APIModelLoader
from .decorators import validate_model, model_update_notification
from .config import ModelManagerConfig

from intelliw.utils.logger import _get_framework_logger

# 配置日志
logger = _get_framework_logger()

@singleton
class ModelManager(BaseModelManager):
    def __init__(self, initial_model_local_path, load_func):
        """初始化模型管理器"""
        logger.info('Initializing ModelManager...')
        self.load_func = load_func
        self.running = False
        self.update_thread = None
        self.model_lock = threading.RLock()

        # 初始化配置管理器
        self.config_manager = ModelManagerConfig()
        self.update_interval = self.config_manager.get('MODEL_UPDATE_INTERVAL')
        self.model_id = self.config_manager.get('TRAININGASSIGNMENT_MODEL_ID')
        self.modelfiles_by_modelid_url = self.config_manager.get('MODELFILES_BY_MODELID_URL')
        self.hot_reload_mode = self.config_manager.get('HOT_RELOAD_MODE')
        
        # 双缓冲模型存储优化：使用list[2]数组和curindex
        self.model_buffer = [None, None]  # 存储两个模型的缓冲区
        self.curindex = 0  # 当前活跃模型的索引
        # 初始化模型存储路径
        self.inited_model_local_path = initial_model_local_path
        self.model_version = None
        self.model_path = None
        self.model_modified_date = None

        logger.info(f'Using TRAINING  MODEL ID: {self.model_id}')

        # 初始化模型加载器
        self.model_loader = APIModelLoader(self.config_manager, load_func=self.load_func)

        # 初始加载模型
        self._initial_load_model()

        # 启动定时更新任务
        if self.hot_reload_mode:
            self._start_update_task()

    def _initial_load_model(self):
        """初始加载模型"""
        try:
            logger.info('Initial loading model...')

            # 配置初始模型版本
            model_path = self.inited_model_local_path
            version = os.environ.get('MODEL_VERSION', 0)
            model_modified_date = time.strftime("%Y-%m-%d %H:%M:%S", time.localtime())

            # 加载模型
            model_data = self._load_local_model(model_path)
            if model_data:
                with self.model_lock:
                    self.model_buffer[self.curindex] = model_data
                    self.model_version = version
                    self.config_manager.set('CURRENT_MODEL_VERSION', version)
                    self.config_manager.set('CURRENT_MODEL_MODIFIED_DATE', model_modified_date)
                logger.info(f'Successfully loaded initial model version: {version}, modified_date: {model_modified_date}')
                # 上报状态
                self._report_status('loaded')
            else:
                logger.error('Failed to download initial model')
        except Exception as e:
            logger.error(f'Error during initial model loading: {str(e)}', exc_info=True)

    def _load_new_model(self, model_path: str, version: str, modified_date: str) -> bool:
        """加载新模型到备用缓冲区"""
        try:
            model_data = self._download_and_prepare_model(model_path, version)
            if not model_data:
                return False

            with self.model_lock:
                # 计算缓冲区索引
                old_index = self.curindex
                new_index = (self.curindex + 1) % 2
                # 加载到备用缓冲区
                self.model_buffer[new_index] = model_data
                # 切换缓冲区索引
                self.curindex = new_index
                self.model_version = version
                self.model_modified_date = modified_date
                self.config_manager.set('CURRENT_MODEL_VERSION', version)
                self.config_manager.set('CURRENT_MODEL_MODIFIED_DATE', modified_date)
                time.sleep(5)
                # 释放旧模型内存
                self.model_buffer[old_index] = None
            logger.info(f'Successfully updated to model version: {version}')
            # 上报状态
            self._report_status('updated')
            return True
        except Exception as e:
            logger.error(f'Error loading new model: {str(e)}', exc_info=True)
            return False

    def get_active_model(self) -> Optional[Any]:
        """获取当前活跃模型"""
        # with self.model_lock:
        return self.model_buffer[self.curindex]

    @validate_model
    def _download_and_prepare_model(self, model_path: str, version: str) -> Optional[Any]:
        """下载并准备模型"""
        model_data = self.model_loader.load(model_path, version)
        if model_data:
            # 返回模型
            return model_data
        return None

    @validate_model
    def _load_local_model(self, local_model_path: str) -> Optional[Any]:
        """加载本地模型"""
        model_data = self.model_loader.load_model(local_model_path)
        if model_data:
            # 返回模型
            return model_data
        return None

    def _report_status(self, status: str) -> None:
        """向AI工作坊上报状态变更"""
        try:
            logger.info(f'Reporting status: {status}')
            # 使用本地定义的装饰器替代缺失的al_decorator模块
            from .decorators import train_info_report

            @train_info_report
            def report_status_impl(model_id, version, status):
                # 实际的上报逻辑
                return True

            report_status_impl(self.model_id, self.model_version, status)
            logger.info('Status reported successfully')
        except Exception as e:
            logger.error(f'Error reporting status: {str(e)}', exc_info=True)

    @model_update_notification
    def _check_for_updates(self):
        """检查模型更新"""
        try:
            logger.info('Checking for model updates...')
            update_info = self.model_loader.check_for_updates()

            # 检查是否有更新
            if update_info:
                logger.info(f'New model version available: {update_info["version"]}, last modified: {update_info["modified_date"]}')
                model_path = update_info['path']
                self._load_new_model(model_path, update_info['version'], update_info["modified_date"])
                return update_info
            else:
                logger.info(f'No new updates. Current version: {self.model_version}, modified_date: {self.model_modified_date}')
                return None
        except Exception as e:
            logger.error(f'Error checking for updates: {str(e)}', exc_info=True)
            return None



    def _update_task(self):
        """定时更新任务"""
        while self.running:
            self._check_for_updates()
            # 等待指定间隔
            for _ in range(self.update_interval):
                if not self.running:
                    break
                time.sleep(1)

    def _start_update_task(self):
        """启动定时更新任务"""
        if self.running:
            logger.warning('Update task is already running')
            return

        self.running = True
        self.update_thread = threading.Thread(target=self._update_task, daemon=True)
        self.update_thread.start()
        logger.info('Update task started')

    def get_model_version(self) -> str:
        """获取当前模型版本"""
        return self.model_version

    def get_model(self):
        return self.get_active_model()

    def load_model(self, model_path: str, version: str) -> bool:
        """加载指定路径的模型"""
        return self._load_new_model(model_path, version)

    def check_for_updates(self) -> bool:
        """检查模型更新"""
        update_info = self._check_for_updates()
        return update_info is not None

    def start(self) -> None:
        """启动模型管理器"""
        self._start_update_task()

    def stop(self):
        """停止模型管理器"""
        logger.info('Stopping ModelManager...')
        self.running = False
        if self.update_thread and self.update_thread.is_alive():
            self.update_thread.join(timeout=5)
        logger.info('ModelManager stopped')

    def set_config(self, config: Dict[str, Any]) -> None:
        """设置配置"""
        for key, value in config.items():
            self.config_manager.set(key, value)

        # 更新相关属性
        self.update_interval = self.config_manager.get('MODEL_UPDATE_INTERVAL')
        self.model_id = self.config_manager.get('TRAININGASSIGNMENT_MODEL_ID')

    def get_config(self) -> Dict[str, Any]:
        """获取配置"""
        return self.config_manager.to_dict()


# 单例模式
_model_manager_instance = None

def get_model_manager(initial_model_local_path, load_func) -> ModelManager:
    global _model_manager_instance
    if _model_manager_instance is None:
        _model_manager_instance = ModelManager(initial_model_local_path, load_func)
    return _model_manager_instance


# 确保在导入时就创建实例
# model_manager = get_model_manager()

# if __name__ == '__main__':
#     # 测试代码
#     model_manager = get_model_manager()
#     try:
#         # 保持程序运行一段时间以观察效果
#         for _ in range(60):
#             print(f'Current model version: {model_manager.get_model_version()}')
#             time.sleep(1)
#     finally:
#         model_manager.stop()


'''
Author: hexu
Date: 2021-10-14 14:54:05
LastEditTime: 2023-05-23 18:46:45
LastEditors: Hexu
Description: nlp数据集
FilePath: /iw-algo-fx/intelliw/datasets/datasource_nlp_corpora.py
'''
from random import Random
import shutil
import time
import os
import fileinput
import zipfile
from intelliw.datasets.datasource_base import AbstractDataSource, DataSourceType, DatasetSelector
from intelliw.utils import iuap_request
from intelliw.utils.logger import _get_framework_logger
import traceback
from intelliw.utils.storage_service import StorageService
from intelliw.config import config
from intelliw.utils.exception import DatasetException
from intelliw.utils import unzip_file
from intelliw.datasets.spliter import get_set_count, check_split_ratio

logger = _get_framework_logger()


class CorpusType:
    csv = 'csv'
    json = 'json'
    txt = 'txt'

    ItoN = {20: "txt", 21: "csv", 22: "json"}


class CorpusInputType:
    local = 'local'
    file = 'file'
    row = 'row'


class DataSourceNLPCorpora(AbstractDataSource):
    """
    nlp语料处理
    """
    # dataset_class 一共有 训练 验证 测试 三种数据集
    dataset_class = 3
    # file_index 文件名称，self._file_batch行拆分一个
    file_index = [1] * dataset_class
    # 共几个数据集， 当大于1时，写到同一个文件
    dataset_count = 0
    # counts 数据集总行数
    counts = [0] * dataset_class

    def __init__(self, source_address: str, ds_id: str, source_train_type: int, tenant_id: str = None,
                 reader_type: str = CorpusInputType.local, download_dir='./tmp_local_nlp_corpora_origin_data/'):
        DataSourceNLPCorpora.dataset_count += 1
        self.dataset_no = DataSourceNLPCorpora.dataset_count

        self.ratio = [config.TRAIN_DATASET_RATIO,
                      config.VALID_DATASET_RATIO, config.TEST_DATASET_RATIO]
        check_split_ratio(*self.ratio)

        self._file_batch = 5e4
        self.random = Random(config.DATA_RANDOM_SEED)
        self.download_dir = download_dir

        self.ds_id = ds_id
        self.source_address = source_address
        self.source_train_type = source_train_type
        self._file_suffix = self._get_file_suffix(self.source_train_type)
        self.reader_type = reader_type if reader_type else CorpusInputType.local
        self.dirpath = [config.NLP_CORPORA_TRAIN_FILEPATH,
                        config.NLP_CORPORA_VAL_FILEPATH, config.NLP_CORPORA_TEST_FILEPATH]
        self.tenant_id = tenant_id if tenant_id is not None else config.TENANT_ID

        # 3种模式  local 和 file 需要进行预处理
        # local  语料文件在本地 本地测试用
        # file   接口获取的为语料文件
        # row    接口获取的为某行语料
        self._csv_label = None
        if self.reader_type != CorpusInputType.row:
            if self.reader_type == CorpusInputType.local:
                local_corpus = self.source_address
            else:
                local_corpus = self.__download_file(
                    self.source_address, self.ds_id, self.source_train_type, download_dir)
            self._file_list, self._csv_label = self._file_process(local_corpus)

    def total(self):
        return 1

    def reader(self, *args, **kwargs):
        return list()

    def corpora_process(self, split_mode: int = 0):
        """nlp语料切分

        Args:
            split_mode (int, optional): 0顺序 1乱序. Defaults to 0.

        此方法中所有长度3的列表, 均代表 [训练, 验证, 测试]
        """
        # 生成数据文件夹
        self._gen_corpora_dir()

        start_time = time.time()
        # 切分函数的缓存, 减少切分次数
        spliter_cache = {}
        # io_reader 数据集写入文件io
        io_reader = []
        for i in range(DataSourceNLPCorpora.dataset_class):
            file_path = os.path.join(
                self.dirpath[i], f"{DataSourceNLPCorpora.file_index[i]}.{self._file_suffix}")
            io_reader.append(open(file_path, "a", encoding="utf-8"))
            if self._csv_label:
                io_reader[i].write(self._csv_label)

        # 进行多文件循环数据集拆分
        for epoch, part in enumerate(self._get_reader()):
            # split_mode 根据拆分模式进行处理
            if split_mode:
                self.random.shuffle(part)
            part_count = len(part)

            # set_data 此part数据集原始数据
            set_data = [[]] * DataSourceNLPCorpora.dataset_class
            nums = spliter_cache.get(part_count)
            if nums is None:
                nums = [part_count, 0, 0] if part_count < 10 else get_set_count(
                    part_count, *self.ratio)
                spliter_cache[part_count] = nums

            _start, _end = 0, 0
            for idx, num in enumerate(nums):
                _end += num
                set_data[idx] = part[_start:] if idx == len(
                    nums) - 1 else part[_start:_end]
                _start += num
            part.clear()

            # 更新数据集总数据
            for idx, num in enumerate(nums):
                DataSourceNLPCorpora.counts[idx] += num

            # 将数据写入对应文件
            [i.writelines(s) for i, s in zip(io_reader, set_data)]

            # 5轮刷新一次缓存区，减少内存压力
            if epoch and epoch % 5 == 0:
                [i.flush() for i in io_reader]

            # 根据数据条数,更新句柄
            for idx, count in enumerate(DataSourceNLPCorpora.counts):
                if count > self._file_batch * DataSourceNLPCorpora.file_index[idx]:
                    DataSourceNLPCorpora.file_index[idx] += 1
                    file_path = os.path.join(
                        self.dirpath[idx],
                        f"{DataSourceNLPCorpora.file_index[idx]}.{self._file_suffix}"
                    )
                    io_reader[idx].close()
                    io_reader[idx] = open(file_path, "a", encoding="utf-8")
                    if self._csv_label:
                        io_reader[idx].write(self._csv_label)

            logger.info(
                f"语料文件下载中: train: {DataSourceNLPCorpora.counts[0]} 条, validation: {DataSourceNLPCorpora.counts[1]} 条, test: {DataSourceNLPCorpora.counts[2]} 条")

        # 关闭所有句柄
        [io.close() for io in io_reader]
        logger.info(
            f"语料[{self.dataset_no}]下载完成: total: {sum(DataSourceNLPCorpora.counts)} 条, train: {DataSourceNLPCorpora.counts[0]} 条, validation: {DataSourceNLPCorpora.counts[1]} 条, test: {DataSourceNLPCorpora.counts[2]} 条, time: {time.time() - start_time}")

        # 清空下载文件夹
        shutil.rmtree(self.download_dir, ignore_errors=True)

    def _gen_corpora_dir(self):
        logger = _get_framework_logger()
        filepath = os.path.join('./', config.NLP_CORPORA_FILEPATH)
        if os.path.exists(filepath) and self.dataset_no == 1:
            logger.warn(f"语料数据保存路径存在:{filepath}, 正在删除路径内容")
            shutil.rmtree(filepath, ignore_errors=True)
        os.makedirs(config.NLP_CORPORA_TRAIN_FILEPATH, 0o755, True)
        os.makedirs(config.NLP_CORPORA_VAL_FILEPATH, 0o755, True)
        os.makedirs(config.NLP_CORPORA_TEST_FILEPATH, 0o755, True)

    def __download_file(self, source_address, ds_id, source_train_type, download_dir):
        start_time = time.time()
        logger.info('Downloading nlp corpora from %s to %s',
                    source_address, download_dir)

        filepath = os.path.join('./', download_dir)
        if os.path.exists(filepath):
            shutil.rmtree(filepath, ignore_errors=True)
        os.makedirs(download_dir, 0o755, True)

        # 1 获取s3链接
        request_data = {'dsIds': ds_id,
                        'type': source_train_type, 'tenantId': self.tenant_id}
        response = iuap_request.post_json(
            url=source_address, json=request_data, timeout=config.DATA_SOURCE_READ_TIMEOUT)
        response.raise_for_status()
        file_links = response.json.get("data", [])  # type: ignore
        if len(file_links) == 0:
            raise DatasetException(
                f"nlp corpora is empty, get data is {response.json}")

        # 2 下载文件
        for idx, link in enumerate(file_links):
            filename = f"{idx + 1}.{self._file_suffix}"
            filepath = os.path.join(
                download_dir, f"{idx + 1}.{self._file_suffix}")
            try:
                downloader = StorageService(
                    link, "download")
                downloader.download(filepath, stream=True)
                logger.info(
                    f"NLP语料 {filename} 下载成功, 耗时:{time.time() - start_time}s")
            except Exception as e:
                err = traceback.format_exc()
                raise DatasetException(f"NLP语料下载失败: {err}")
        return download_dir

    def _file_process(self, path):
        if zipfile.is_zipfile(path):
            logger.info("解压语料文件")
            dirpath = unzip_file(path)
        elif os.path.isdir(path):
            dirpath = path
        else:
            raise DatasetException(f"不支持的文件格式{path}")
        # 读取文件夹， 筛选文件
        file_list = []

        def get_file(d, f):
            for i in os.listdir(d):
                p = os.path.join(d, i)
                if i.endswith(self._file_suffix):
                    f.append(p)
                elif os.path.isdir(p):
                    get_file(p, f)
            return f

        file_list = get_file(dirpath, file_list)

        if len(file_list) == 0:
            raise DatasetException(
                f"未获取语料文件， 文件格式应为 .{self._file_suffix} 后缀文件")

        csv_label = self._get_csv_label(file_list)
        return file_list, csv_label

    def _get_csv_label(self, file_list):
        # 获取csv表头
        if self._file_suffix == CorpusType.csv:
            fp = open(file_list[0], encoding='utf-8')
            csv_label = fp.__next__()
            fp.close()
            return csv_label
        else:
            return None

    def _get_file_suffix(self, source_train_type):
        if source_train_type not in CorpusType.ItoN.keys():
            raise DatasetException(
                f"NLP语料类型错误: CorpusType is {source_train_type}, 20-txt 21-csv 22-json")
        return CorpusType.ItoN[source_train_type]

    def __call__(self):
        abspath = os.path.abspath('.')
        return {'path': os.path.join(abspath, config.NLP_CORPORA_FILEPATH),
                'train_set': os.path.join(abspath, config.NLP_CORPORA_TRAIN_FILEPATH),
                'val_set': os.path.join(abspath, config.NLP_CORPORA_VAL_FILEPATH),
                'test_set': os.path.join(abspath, config.NLP_CORPORA_TEST_FILEPATH)}

    # reader
    def _get_reader(self, pagesize=5000):
        if self.reader_type == "row":
            return DataSourceNLPCorpora.__OnlineReader(self.source_address, self.ds_id, self.tenant_id, pagesize)
        else:
            return self.__local_reader(self._file_list, pagesize)

    def __local_reader(self, file_paths, limit=5000, encoding="utf-8"):
        hook = fileinput.hook_encoded(encoding=encoding)
        with fileinput.input(files=file_paths, openhook=hook) as f:
            while True:
                part = []
                for _ in range(limit):
                    try:
                        t = f.__next__()
                        if f.isfirstline() and self._file_suffix == CorpusType.csv:
                            continue
                        part.append(t)
                    except StopIteration:
                        if part:
                            break
                        else:
                            return
                yield part

    class __OnlineReader:
        def __init__(self, input_address, ds_id, tenant_id, limit=5000):
            """
            eg. 91 elements, page_size = 20, 5 pages as below:
            [0,19][20,39][40,59][60,79][80,90]
            offset 15, limit 30:
            [15,19][20,39][40,44]
            offset 10 limit 5:
            [10,14]
            """
            self.logger = _get_framework_logger()
            self.input_address = input_address
            self.ds_id = ds_id
            self.page_size = max(100, limit)
            self.page_num = 0
            self.total_read = 0
            self.tenant_id = tenant_id

        @property
        def iterable(self):
            return True

        def __iter__(self):
            return self

        def __next__(self):
            """
            {
                "data": {
                    "content": ["...","..."],
                    "pageNumber": 0,
                    "pageSize": 1,
                    "totalElements": 14,
                    "totalPages": 14
                },
                "msg": "成功",
                "status": 1
            }
            """
            page = None
            try:
                page = self._read_page(self.page_num, self.page_size)
                assert page is not None, "获取的数据为空"
                if page['status'] != 1:
                    raise Exception(page['msg'])
                if page['data']['totalElements'] == 0:
                    raise Exception("nlp 数据获取失败, 无已标注语料")
                content = page['data']['content']
                content = [i if i.endswith(
                    "\n") else f"{i}\n" for i in content]
            except Exception as e:
                errmsg = f"nlp 数据获取失败, \nds_id: [{self.ds_id}], \nerror: [{e}], \nresponse: {page}"
                raise DatasetException(errmsg)

            if len(content) == 0:
                raise StopIteration
            self.page_num += 1
            self.total_read += len(content)
            return content

        def _read_page(self, page_num, page_size):
            """
            调用智能分析接口，分页读取数据
            :param page_num: 页码，从 0 开始
            :param page_size:  每页大小
            :return:
            """
            request_data = {'dsId': self.ds_id, 'pageNumber': page_num,
                            'pageSize': page_size, 'yTenantId': self.tenant_id}
            response = iuap_request.post_json(
                url=self.input_address, json=request_data, timeout=config.DATA_SOURCE_READ_TIMEOUT)
            response.raise_for_status()
            return response.json


DatasetSelector.register_func(DataSourceType.NLP_CORPORA, DataSourceNLPCorpora, {
    "source_address": ["NLP_CORPORA_PATH", "INPUT_ADDR"],
    "ds_id": "INPUT_DATA_SOURCE_ID",
    "source_train_type": "INPUT_DATA_SOURCE_TRAIN_TYPE",
    "reader_type": "NLP_CORPORA_INPUT_TYPE",
    "tenant_id": "TENANT_ID"})



'''
Author: Hexu
Date: 2022-08-16 10:23:16
LastEditors: Hexu
LastEditTime: 2023-03-28 16:50:55
FilePath: /iw-algo-fx/intelliw/utils/spark_process/spark.py
Description: spark function
'''
import os
from intelliw.utils.exception import DatasetException
from pyspark import SparkContext
from pyspark import pandas as ps
from pyspark.sql import SparkSession
from pyspark.sql.dataframe import DataFrame as SparkDF
import warnings
warnings.filterwarnings('ignore')


app_name = "INTELLIW_SPARK_NAME"


class ReadFileType:
    CSV = "csv"
    TXT = "txt"
    JSON = "json"


class Spark(object):
    """
    Container class holding SparkContext and SparkSession instances, so that any changes
    will be propagated across the application
    """
    sc = None
    spark = None
    ps = ps

    # config

    def __new__(cls, *args, **kwargs):
        if cls is Spark:
            raise TypeError(
                'SparkSession & SparkContext container class may not be instantiated.')
        return object.__new__(cls, *args, **kwargs)

    @classmethod
    def get_sc(cls, master=None, appName=None, sparkHome=None, pyFiles=None, environment=None,
               batchSize=0, conf=None, gateway=None, jsc=None):
        """Creates and initializes a new `SparkContext` (the old one will be stopped).
        Argument signature is copied from `pyspark.SparkContext
        <https://spark.apache.org/docs/latest/api/python/pyspark.html#pyspark.SparkContext>`_.
        """
        if cls.sc is not None:
            cls.sc.stop()
        cls.sc = SparkContext(master=master, appName=appName, sparkHome=sparkHome, pyFiles=pyFiles, environment=environment, batchSize=batchSize,
                              conf=conf, gateway=gateway, jsc=jsc)
        cls.__init_spark()

    @classmethod
    def get_spark(cls, master=None, appName=None, conf=None, hive_support=False):
        """Creates and initializes a new `SparkSession`. Argument signature is copied from
        `pyspark.sql.SparkSession
        <https://spark.apache.org/docs/latest/api/python/pyspark.sql.html#pyspark.sql.SparkSession>`_.
        """
        if cls.spark is not None:
            return Spark

        sess = SparkSession.builder
        if master:
            sess.master(master)

        if appName:
            sess.appName(appName)
        else:
            sess.appName(app_name)

        if conf:
            conf["spark.sql.debug.maxToStringFields"] = 10000
            sess.config(conf=conf)
        else:
            sess.config("spark.sql.debug.maxToStringFields", "10000")

        if hive_support:
            sess.enableHiveSupport()

        cls.spark = sess.getOrCreate()
        return Spark

    @classmethod
    def init_default(cls):
        """Create and initialize a default SparkContext and SparkSession"""
        if cls.spark is None and cls.sc is None:
            cls.__init_sc()
            cls.__init_spark()
        return Spark

    @classmethod
    def __init_sc(cls):
        cls.sc = SparkContext(appName=app_name).getOrCreate()

    @classmethod
    def __init_spark(cls):
        cls.spark = SparkSession.builder.appName(app_name).config(
            "spark.sql.debug.maxToStringFields", "10000").getOrCreate()

    @classmethod
    def read_file(cls, filetype, filepath) -> SparkDF:
        if not os.path.isdir(filepath) and \
           not os.path.isfile(filepath) and \
           not os.path.islink(filepath):
            raise DatasetException(
                "spark read file error: Not a file、directory or hdfs link")
        if filetype == ReadFileType.CSV:
            return cls.spark.read.option("inferSchema", "true") \
                .option("header", "true") \
                .option("encoding", "gbk") \
                .csv(filepath)
        elif filetype == ReadFileType.TXT:
            return cls.spark.read.text(filepath)
        elif filetype == ReadFileType.JSON:
            return cls.spark.read.json(filepath)
        else:
            raise DatasetException("An unsupported spark type")

    @staticmethod
    def save_file(data, filetype, filepath) -> None:
        if filetype == ReadFileType.CSV:
            data.repartition(1) \
                .write.mode("overwrite") \
                .option("mapreduce.fileoutputcommitter.marksuccessfuljobs", "false") \
                .option("header", "true") \
                .option("delimiter", ",") \
                .csv(filepath)



import os
import time
import threading
import unittest
from unittest.mock import patch, MagicMock, Mock
import tempfile
import json

# 导入重构后的模型管理器和相关模块
from .model_manager import get_model_manager, ModelManager
from .model_loader import APIModelLoader
from .config import ConfigManager, ModelManagerConfig


class TestModelManager(unittest.TestCase):
    def setUp(self):
        # 保存原始环境变量
        self.original_env = os.environ.copy()
        # 设置测试环境变量
        os.environ['MODEL_ID'] = 'test_model'
        os.environ['MODEL_UPDATE_INTERVAL'] = '10'
        os.environ['CONSOLE_API_URL'] = 'http://test-console-api.com'
        os.environ['OSS_API_URL'] = 'http://test-oss-api.com'
        os.environ['MODEL_LOAD_TIMEOUT'] = '5'

        # 创建临时配置文件用于测试
        self.temp_config_dir = tempfile.TemporaryDirectory()
        self.temp_config_path = os.path.join(self.temp_config_dir.name, 'test_config.json')
        with open(self.temp_config_path, 'w') as f:
            json.dump({
                'MODEL_ID': 'test_model_from_file',
                'MODEL_UPDATE_INTERVAL': 5
            }, f)

        # 重置模型管理器单例实例
        from .model_manager import _model_manager_instance
        _model_manager_instance = None

        # 保存原始的单例装饰器
        from intelliw.utils.modelmanager.singleton import singleton
        self.original_singleton = singleton

        # 定义一个临时的非单例装饰器
        def non_singleton(cls):
            return cls

        # 替换单例装饰器
        import intelliw.utils.modelmanager.singleton
        intelliw.utils.modelmanager.singleton.singleton = non_singleton

        # 重新加载ModelManager模块以应用新的装饰器
        import importlib
        import intelliw.utils.modelmanager.model_manager
        importlib.reload(intelliw.utils.modelmanager.model_manager)

        global ModelManager, get_model_manager
        from intelliw.utils.modelmanager.model_manager import ModelManager, get_model_manager

        # 模拟配置管理器，使用ModelManagerConfig作为spec因为它实现了to_dict方法
        self.mock_config_manager = Mock(spec=ModelManagerConfig)
        self.mock_config_manager.get.side_effect = lambda key: {
            'TRAININGASSIGNMENT_MODEL_ID': 'test_model',
            'MODEL_UPDATE_INTERVAL': 10,
            'MODELFILES_BY_MODELID_URL': 'http://test-console-api.com',
            'MODEL_LOAD_TIMEOUT': 5,
            'REQUEST_TIMEOUT': 10,
            'MODEL_UPDATE_TIMEOUT': 10,
            'CURRENT_MODEL_VERSION': None
        }.get(key)
        self.mock_config_manager.to_dict.return_value = {
            'TRAININGASSIGNMENT_MODEL_ID': 'test_model',
            'MODEL_UPDATE_INTERVAL': 10,
            'CONSOLE_API_URL': 'http://test-console-api.com',
            'OSS_API_URL': 'http://test-oss-api.com',
            'MODEL_LOAD_TIMEOUT': 5
        }

        # 模拟模型加载器
        self.mock_model_loader = Mock(spec=APIModelLoader)
        self.model_data = {'weights': [1.0, 2.0, 3.0], 'biases': [0.1, 0.2]}
        self.mock_model_loader.check_for_updates.return_value = {
            'path': 'test_model_path',
            'version': 'v1.0'
        }
        self.mock_model_loader.load.return_value = self.model_data

    def tearDown(self):
        # 恢复原始环境变量
        os.environ.clear()
        os.environ.update(self.original_env)

        # 清理临时目录
        self.temp_config_dir.cleanup()

        # 恢复原始的单例装饰器
        import intelliw.utils.modelmanager.singleton
        intelliw.utils.modelmanager.singleton.singleton = self.original_singleton

        # 重新加载ModelManager模块以应用原始装饰器
        import importlib
        import intelliw.utils.modelmanager.model_manager
        importlib.reload(intelliw.utils.modelmanager.model_manager)
        # 停止模型管理器
        try:
            model_manager = get_model_manager()
            model_manager.stop()
        except:
            pass

    def test_singleton(self):
        """测试单例模式"""
        manager1 = get_model_manager(None)
        manager2 = get_model_manager(None)
        self.assertIs(manager1, manager2)

    def test_initial_load_model(self):
        """测试初始加载模型"""
        # 创建模拟配置管理器
        mock_config = MagicMock()
        mock_config.get.side_effect = lambda key: {
            'TRAININGASSIGNMENT_MODEL_ID': 'test_model',
            'MODEL_UPDATE_INTERVAL': 10
        }.get(key)

        # 创建模拟模型加载器
        mock_model_loader = MagicMock()

        # 模拟check_for_updates返回有效的更新信息
        update_info = {
            'path': 'test_model_path',
            'version': 'v1.0'
        }
        mock_model_loader.check_for_updates.return_value = update_info

        # 模拟load返回原始数据
        raw_model_data = {'test': 'model'}
        mock_model_loader.load.return_value = raw_model_data

        # 直接创建ModelManager实例，绕过单例和构造函数
        model_manager = object.__new__(ModelManager)

        # 手动设置实例属性
        model_manager.model_buffer = [None, None]
        model_manager.curindex = 0
        model_manager.model_version = None
        model_manager.model_lock = threading.RLock()
        model_manager.running = False
        model_manager.update_thread = None
        model_manager.config_manager = mock_config
        model_manager.update_interval = 10
        model_manager.model_id = 'test_model'
        model_manager.model_loader = mock_model_loader

        # 只调用_initial_load_model进行测试
        model_manager._initial_load_model()

        # 验证模型加载器方法被正确调用
        mock_model_loader.check_for_updates.assert_called_once()
        mock_model_loader.load.assert_called_once_with(update_info['path'])

        # 验证模型版本和数据
        self.assertEqual(model_manager.get_model_version(), 'v1.0')
        
        # 验证模型数据经过了prepare_model_params处理
        from intelliw.utils.modelmanager.utils import prepare_model_params
        processed_data = prepare_model_params(raw_model_data)
        self.assertEqual(model_manager.get_active_model(), processed_data)

        # 单独测试_start_update_task
        model_manager._start_update_task()
        self.assertTrue(model_manager.running)
        self.assertIsNotNone(model_manager.update_thread)
        self.assertTrue(model_manager.update_thread.is_alive())

    def test_check_for_updates_no_update(self):
        """测试检查更新（无更新）"""
        # 配置mock返回值
        self.mock_model_loader.check_for_updates.return_value = {
            'path': 'test_model_path',
            'version': 'v1.0'
        }

        # 初始化模型管理器
        model_manager = get_model_manager(None)
        model_manager.model_version = 'v1.0'  # 手动设置当前版本
        model_manager.model_loader = self.mock_model_loader

        # 调用检查更新方法
        result = model_manager._check_for_updates()

        # 验证没有更新
        self.assertEqual(model_manager.get_model_version(), 'v1.0')
        self.assertIsNone(result)

    def test_check_for_updates_with_update(self):
        """测试检查更新（有更新）"""
        # 配置mock返回值
        self.mock_model_loader.check_for_updates.return_value = {
            'path': 'new_model_path',
            'version': 'v2.0'
        }
        new_model_data = {'weights': [4.0, 5.0, 6.0], 'biases': [0.3, 0.4]}
        self.mock_model_loader.load.return_value = new_model_data

        # 初始化模型管理器
        model_manager = get_model_manager(None)
        model_manager.model_version = 'v1.0'  # 手动设置当前版本
        model_manager.model_buffer = [{'old': 'model'}, None]
        model_manager.curindex = 0
        model_manager.model_loader = self.mock_model_loader

        # 调用检查更新方法
        result = model_manager._check_for_updates()

        # 验证更新成功
        self.assertEqual(model_manager.get_model_version(), 'v2.0')
        self.assertEqual(model_manager.get_active_model(), new_model_data)
        self.assertEqual(result['version'], 'v2.0')

    def test_stop(self):
        """测试停止模型管理器"""
        # 创建一个ModelManager实例
        model_manager = ModelManager(None)
        model_manager.model_loader = self.mock_model_loader
        model_manager.config_manager = self.mock_config_manager
        
        # 启动更新任务
        model_manager._start_update_task()
        
        # 确保线程已启动
        time.sleep(0.1)
        
        # 停止模型管理器
        model_manager.stop()
        
        # 等待线程结束
        if model_manager.update_thread:
            model_manager.update_thread.join(timeout=5)  # 设置超时以避免无限等待
        
        # 验证线程已停止
        self.assertFalse(model_manager.running)
        if model_manager.update_thread:
            self.assertFalse(model_manager.update_thread.is_alive())

    def test_config_management(self):
        """测试配置管理功能"""
        model_manager = get_model_manager(None)
        model_manager.config_manager = self.mock_config_manager

        # 测试获取配置
        self.assertEqual(model_manager.get_config()['TRAININGASSIGNMENT_MODEL_ID'], 'test_model')
        self.assertEqual(model_manager.get_config()['MODEL_UPDATE_INTERVAL'], 10)

        # 测试设置配置
        model_manager.set_config({
            'TRAININGASSIGNMENT_MODEL_ID': 'new_test_model',
            'MODEL_UPDATE_INTERVAL': 30
        })

        # 验证配置更新
        self.mock_config_manager.set.assert_any_call('TRAININGASSIGNMENT_MODEL_ID', 'new_test_model')
        self.mock_config_manager.set.assert_any_call('MODEL_UPDATE_INTERVAL', 30)

    def test_get_model_manager_function(self):
        """测试get_model_manager函数"""
        manager = get_model_manager(None)
        self.assertIsInstance(manager, ModelManager)


if __name__ == '__main__':
    unittest.main()


'''
Author: hexu
Date: 2021-10-25 15:20:34
LastEditTime: 2023-05-26 14:26:04
LastEditors: Hexu
Description: 本地使用和安装包的入口文件
FilePath: /iw-algo-fx/intelliw/interface/entrypoint.py
'''
import os
import zipfile
import shutil
from intelliw.config import config, dataset_config
from absl.flags import argparse_flags as argparse


def __parse_args():
    parser = argparse.ArgumentParser(
        description="Entry files for local debug and pypi packages")

    parser.add_argument("-p", "--path", default=os.getcwd(),
                        type=str, help="project's path, default is 'os.getcwd()'")
    parser.add_argument("-o", "--output_path", default="intelliw_output_dir",
                        type=str, help="package file output path")
    parser.add_argument("task", nargs="?", default="version", type=str,
                        help="""version: intelliw version | init: initialize the product | init_docs: initialize the document |
                        init_debug: initialize the debug file | import_alg | import_model | train | infer |
                        package_iwa | package_iwm | package_iwp | default is 'import_alg'""")

    init = parser.add_argument_group(
        '[init/init_docs/init_debug]', description='intelliw init -o output_path -n mydemo')
    init.add_argument("-n", "--name", default="example",
                      type=str, help="algorithms name")

    train = parser.add_argument_group(
        '[train]', description='intelliw train -p project_path -r set_ratio')
    train.add_argument("-c", "--csv", default=None,
                       type=str, help="path of csv files as training data")
    train.add_argument("-C", "--corpus", default=None,
                       type=str, help="path of nlp corpus dir as training data")
    train.add_argument('-r', "--set_ratio", default="0.7:0.3:0", type=str,
                       help="train set:valid set:test set ratio, between 0.0 and 1.0, default is 0.7:0.3:0")
    train.add_argument('-t', "--test_size", default=None, type=str,
                       help="test_size, between 0.0 and 1.0, default not use")
    train.add_argument("-d", "--dataset", default="",
                       type=str, help="dataset id by aiworker")

    infer = parser.add_argument_group(
        '[infer]', description='intelliw infer -p project_path -P 8080')
    infer.add_argument("-P", "--port", default=8888, type=int,
                       help="port to listen, default: 8888")

    parser.add_argument_group('[package_iwa/package_iwm]',
                              description='intelliw package_iwa -p project_path -o output_path; package help: zip algorithms to iwa/iwm')
    parser.add_argument_group('[import_alg/import_model]',
                              description='intelliw import_alg -p project_path; import help: test algorithm import')

    return parser.parse_args()


def __args_to_framework_arg(args, path):
    from intelliw.interface import controller
    framework_args = controller.FrameworkArgs()
    framework_args.path = path

    if args.task == 'train':
        framework_args.task = 'train'
    elif args.task == 'infer':
        framework_args.port = args.port

    task_map = {'import_alg': 'importalg', 'import_model': 'importmodel',
                'train': 'train', 'infer': 'apiservice', 'custom': 'custom'}
    framework_args.method = task_map.get(args.task)
    assert framework_args.method is not None, f'failed to execute, unknown task: {args.task}'
    return framework_args


def __package(task: str, alg_path: str, output_path: str):
    """Package files in a directory and save them as a zip file"""

    task_map = {
        'package': '.iwa',
        'package_iwa': '.iwa',
        'package_iwm': '.iwm',
        'package_iwp': '.iwp'
    }
    if task not in task_map:
        raise ValueError(
            f'Unknown package task: {task}, supported tasks are {task_map.keys()}'
        )

    alg_path = os.path.abspath(alg_path)
    dir_name = os.path.basename(alg_path)
    output_path = os.path.abspath(output_path)
    file_name = dir_name + task_map[task]
    output = os.path.join(output_path, file_name)

    # 删除存在的文件
    if os.path.exists(output):
        os.remove(output)
    # 创建保存的文件夹
    if not os.path.exists(output_path):
        os.makedirs(output_path)

    with zipfile.ZipFile(output, 'w', zipfile.ZIP_DEFLATED) as zipf:
        for root, _, files in os.walk(alg_path):
            if any(root.find(x) >= 0 for x in ('__pycache__', '.git', '.idea')):
                continue
            relative_path = root.replace(alg_path, "")
            if relative_path.startswith(os.sep):
                relative_path = relative_path[len(os.sep):]
            output_relative_path = os.path.join(dir_name, relative_path)
            for file in files:
                if file.startswith('.DS_Store'):
                    continue
                print(os.path.join(root, file), " to zip file:",
                      os.path.join(output_relative_path, file))
                zipf.write(os.path.join(root, file),
                           os.path.join(output_relative_path, file))


def __init_algorithms(name: str, output_path: str, task: str):
    package_path = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))

    def alter(file, old_str, new_str):
        file_data = ""
        with open(file, "r", encoding="utf-8") as f:
            for line in f:
                if old_str in line:
                    line = line.replace(old_str, new_str)
                file_data += line
        with open(file, "w", encoding="utf-8") as f:
            f.write(file_data)

    if output_path == "intelliw_output_dir":
        output_path = os.getcwd()

    if not os.path.exists(output_path):
        raise FileNotFoundError(f"No such file or directory: '{output_path}'")

    if task == "init_docs":
        docs_dir = os.path.join(package_path, "docs")
        shutil.copytree(docs_dir, os.path.join(output_path, "docs"))
    if task == "init_debug":
        debug_file = os.path.join(package_path, "utils", "debug_controller.py")
        shutil.copyfile(debug_file, os.path.join(
            output_path, "debug_controller.py"))
    else:
        output_path = os.path.join(output_path, name)
        algorithms_dir = os.path.join(package_path, "algorithms_demo")
        algorithm_yaml = os.path.join(output_path, "algorithm.yaml")
        shutil.copytree(algorithms_dir, output_path)

        docs_dir = os.path.join(package_path, "docs")
        output_docs_dir = os.path.join(output_path, "docs")
        shutil.copytree(docs_dir, output_docs_dir)

        debug_file = os.path.join(package_path, "utils", "debug_controller.py")
        output_debug_file = os.path.join(output_path, "debug_controller.py")
        shutil.copyfile(debug_file, output_debug_file)

        alter(algorithm_yaml, "AlgorithmName", name)

        pycache = os.path.join(name, "__pycache__")
        if os.path.exists(pycache):
            shutil.rmtree(pycache)


def __dataset(dataset_id=None, set_ratio: str=None, test_size=None, local_csv=None, local_corpus=None):
    """
    config dataset
    """
    dataset_ratio = [0.7, 0.3, 0]
    if test_size:
        try:
            ts = float(test_size)
            if ts > 1 or ts < 0:
                raise Exception(
                    f"test_size 比例范围错误:{set_ratio}, 和范围为 [0, 1]")
            dataset_ratio = [1-ts, 0, ts]

        except:
            raise Exception(
                f"test_size 格式错误:{test_size}, 请参考 --test_size 0.3")
    elif set_ratio:
        try:
            tmp_ratios = [float(ratio) for ratio in set_ratio.split(":")]
            if len(tmp_ratios) >=4 or any([ratio < 0 and ratio> 1 for ratio in tmp_ratios]):
                raise Exception(
                    f"数据集比例格式错误:{set_ratio}, 请参考 -r 0.7:0.2:0.1")
            else:
                tmp_sum = sum(tmp_ratios)
                if tmp_sum > 1.0 or tmp_sum < 0:
                    raise Exception(
                        f"数据集比例范围错误:{set_ratio}, 和范围为 [0, 1]")
                if len(tmp_ratios) == 3:   
                    dataset_ratio[0] = tmp_ratios[0]
                    dataset_ratio[1] = tmp_ratios[1]
                    dataset_ratio[2] = tmp_ratios[2]
                elif len(tmp_ratios) == 2:   
                    dataset_ratio[0] = tmp_ratios[0]
                    dataset_ratio[1] = tmp_ratios[1]
                    dataset_ratio[2] = 1-tmp_ratios[0]-tmp_ratios[1]
                elif len(tmp_ratios) == 1:   
                    dataset_ratio[0] = tmp_ratios[0]
                    dataset_ratio[1] = 1-tmp_ratios[0]
                    dataset_ratio[2] = 0
        except:
            raise Exception(
                f"数据集比例格式错误:{set_ratio}, 请参考 -r 0.7:0.2:0.1")
    dataset_config.set_dataset_config(
        dataset_id, dataset_ratio, local_csv, local_corpus)


def run():
    """
    run
    """
    args = __parse_args()

    # version
    if args.task.startswith("version"):
        return

    # init
    if args.task.startswith("init"):
        __init_algorithms(args.name, args.output_path, args.task)
        return

    # package
    if args.task.startswith("package"):
        __package(args.task, args.path, args.output_path)
        return

    # set task
    config.update_by_env()
    import intelliw.interface.controller as controller
    if args.task == 'train':
        from intelliw.utils.gen_model_cfg import generate_model_config_from_algorithm_config as __generate
        # 处理 yaml 文件
        model_yaml_path, algo_yaml_path = os.path.join(
            args.path, 'model.yaml'), os.path.join(args.path, 'algorithm.yaml')

        __generate(algo_yaml_path, model_yaml_path)

        __dataset(args.dataset, args.set_ratio, args.test_size, args.csv, args.corpus)

    framework_args = __args_to_framework_arg(args, args.path)
    config.RUNNING_MODE = 'SCAFFOLDING'
    controller.main(framework_args)


if __name__ == '__main__':
    run()



import io
import os
import sys
import tempfile
import time
from functools import wraps
import cProfile
import pstats

from intelliw.utils.storage_service import StorageService, FileTransferDevice
from intelliw.utils.logger import _get_framework_logger

logger = _get_framework_logger()


def _pre():
    logger.warning("performance profile request...")
    pr = cProfile.Profile()
    pr.enable()
    return pr


def _post(pr):
    pr.disable()
    with io.StringIO() as s:
        sort_by = pstats.SortKey.CUMULATIVE
        ps = pstats.Stats(pr, stream=s).sort_stats(sort_by)
        ps.print_stats()
        profile = s.getvalue()

    logger.warning(f"\n{profile}")

    """
    上传云服务， 通过
    >>> flameprof pipeline.prof > pipeline.svg
    查看火焰图
    """
    with tempfile.NamedTemporaryFile() as tmp:
        tmp_file = tmp.name + '.prof'
        pr.dump_stats(tmp_file)
        fd = FileTransferDevice(
            tmp_file, "profile_result", filename=f"result-{int(time.time())}.prof"
        )
    raw_stats = fd.curkey
    return profile, raw_stats


async def async_performance_profile(func, *args, **kwargs):
    pr = _pre()
    result = await func(*args, **kwargs)
    profile, raw_stats = _post(pr)
    return {"result": result, "profile": profile, "raw_stats": raw_stats}


def performance_profile(func, *args, **kwargs):
    pr = _pre()
    result = func(*args, **kwargs)
    profile, raw_stats = _post(pr)
    return {"result": result, "profile": profile, "raw_stats": raw_stats}


def performance_profile_wrapper(filepath=None):
    def wrapper(function):
        @wraps(function)
        def inner(*args, **kwargs):
            pr = _pre()
            result = function(*args, **kwargs)
            profile, raw_stats = _post(pr)
            return {"result": result, "profile": profile, "raw_stats": raw_stats}

        return inner

    return wrapper



def filter_by_condition(data, cfg):
    """
    Filter data by condition, drop row if conditional column is None.

    condition is in json format:
    {
        "op": [     // 外层条件关系 and 或 or
            "and"
        ],
        "terms": [  // 外层条件 list
            {
                "op": "and", // 内层条件关系 and 或 or
                "items": [   // 内层条件 list
                    {
                        "code": "engName", // 字段 code
                        "col_num": 0,
                        "type": "COMPARE", // 字段类型：STRING,NUMBER,ENUM,DATE,BOOLEAN,COMPARE
                        "op": "gt",        // 操作符：eq-等于 neq-不等于 lt-小于 gt-大于 elt-小于等于 egt-大于等于 leftlike-左包含 rightlike-右包含 like-包含 in-在…之内 nin-不在…之内
                        "value": "123"     // 比较的值
                    }
                ]
            }
        ]
    }

    :param data:
    :param cfg: cfg['condition']
    :return:
    """
    condition_str = cfg['condition']
    import json
    condition = json.loads(condition_str)

    # key: code, val: index
    meta = {}
    for i, item in enumerate(data['meta']):
        meta[item['code']] = i

    result = []
    for row in data['result']:
        if _conditional_column_is_none(meta, row, condition):
            continue
        if _evaluate_row(meta, row, condition):
            result.append(row)

    data['result'] = result
    return data


def _conditional_column_is_none(meta: dict, row: list, condition) -> bool:
    """
    weather conditional column is none

    :return: True if conditional column is None
    """
    terms = condition['terms']
    for term in terms:
        for item in term['items']:
            if row[meta[item['code']]] is None:
                return True
    return False


def _evaluate_row(meta: dict, row: list, condition) -> bool:
    """
    evaluate result_row by condition

    :param meta: key: code, val: index
    :param row: row of data been evaluated by condition
    :param condition:  condition
    :return: True if result_row meets condition or False if not
    """
    terms = condition['terms']
    term_ops = condition['op']
    term_op = None
    current = None
    for i, term in enumerate(terms):
        if term_op is None:
            current = _evaluate_term(term, meta, row)
        elif term_op == 'and':
            current = current and _evaluate_term(term, meta, row)
        elif term_op == 'or':
            current = current or _evaluate_term(term, meta, row)
        else:
            raise ValueError('invalid op: {}'.format(term_op))
        if i < len(term_ops):
            term_op = term_ops[i]

    return current


def _evaluate_term(term, meta, row) -> bool:
    current = None
    op = term['op']
    for item in term['items']:
        if current is None:
            current = _evaluate_item(item, meta, row)
            continue
        if op == 'and':
            current = current and _evaluate_item(item, meta, row)
            if not current:
                return False
        elif op == 'or':
            current = current or _evaluate_item(item, meta, row)
        else:
            raise ValueError('invalid op: {}'.format(op))
    return current


def _evaluate_item(item, meta, row) -> bool:
    op = item['op']
    _type = item['type']
    val = _to_type(_type, row[meta[item['code']]])
    target = item['value']
    # eq-等于 neq-不等于 lt-小于 gt-大于 elt-小于等于 egt-大于等于 leftlike-左包含 rightlike-右包含 like-包含 in-在…之内 nin-不在…之内
    if op == 'eq':
        return val == _to_type(_type, target)
    elif op == 'neq':
        return val != _to_type(_type, target)
    elif op == 'lt':
        return val < _to_type(_type, target)
    elif op == 'gt':
        return val > _to_type(_type, target)
    elif op == 'elt':
        return val <= _to_type(_type, target)
    elif op == 'egt':
        return val >= _to_type(_type, target)
    elif op == 'leftlike':
        val = str(val)
        target = str(target)
        return val.startswith(target)
    elif op == 'rightlike':
        val = str(val)
        target = str(target)
        return val.endswith(target)
    elif op == 'like':
        val = str(val)
        target = str(target)
        return val.__contains__(target)
    elif op == 'in':
        keys = list(map(lambda x: _to_type(_type, x), str(target).split(',')))
        return keys.__contains__(val)
    elif op == 'nin':
        keys = list(map(lambda x: _to_type(_type, x), str(target).split(',')))
        return not keys.__contains__(val)
    elif op == 'null':
        return val is None or val == ''
    elif op == 'notnull':
        return val is not None or val != ''
    else:
        raise ValueError('invalid op: {}'.format(op))


def _to_type(_type, val):
    if _type == 'ENUM':
        return str(val)
    elif _type == 'NUMBER':
        if isinstance(val, int):
            return val
        elif isinstance(val, float):
            return val
        else:
            return eval(str(val))
    elif _type == 'BOOLEAN':
        if isinstance(val, bool):
            return val
        else:
            if str(val) == 'True' or str(val) == 'true' or val == 1 or val == '1':
                return True
            else:
                return False
    else:
        if isinstance(val, str):
            return val
        return str(val) if str(val) else None



'''
Author: Hexu
Date: 2022-05-05 12:07:16
LastEditors: Hexu
LastEditTime: 2022-09-22 19:58:12
FilePath: /iw-algo-fx/intelliw/utils/crontab.py
Description: crontab
'''
import datetime
import types

try:
    from apscheduler.schedulers.blocking import BlockingScheduler
    from apscheduler.schedulers.background import BackgroundScheduler
    from apscheduler.triggers.cron import CronTrigger
except ImportError:
    raise ImportError(
        "\033[31mIf use cronjob, you need: pip install apscheduler\033[0m")


def _convert_special_day(day_of_week):
    """
    将特殊的星期几表示转换为标准表示。
    ? 转换为 *，7 或者 0 转换为 0。
    """
    if day_of_week == "?":
        return "*"
    elif day_of_week == "7":
        return "0"
    return day_of_week


class CrontabError(Exception):
    pass


class MyCronTrigger(CronTrigger):

    @classmethod
    def from_crontab(cls, expr, timezone=None):
        """
        Create a :class:`~CronTrigger` from a standard crontab expression.

        See https://en.wikipedia.org/wiki/Cron for more information on the format accepted here.

        :param expr: minute, hour, day of month, month, day of week
        :param datetime.tzinfo|str timezone: time zone to use for the date/time calculations (
            defaults to scheduler timezone)
        :return: a :class:`~CronTrigger` instance

        """
        if not expr:
            raise ValueError("时间表达式不能为空")
        values = expr.split()

        if len(values) not in [5, 6, 7]:
            raise ValueError(f"时间表达式长度不正确，应为5、6或7，实际为{len(values)}")

        if len(values) == 5:
            return cls(minute=values[0], hour=values[1], day=values[2], month=values[3],
                       day_of_week=_convert_special_day(values[4]), timezone=timezone)
        if len(values) == 6:
            return cls(second=values[0], minute=values[1], hour=values[2], day=values[3], month=values[4],
                       day_of_week=_convert_special_day(values[5]), timezone=timezone)
        if len(values) == 7:
            return cls(second=values[0], minute=values[1], hour=values[2], day=values[3], month=values[4],
                       day_of_week=_convert_special_day(values[5]), year=values[6], timezone=timezone)


class Crontab:
    def __init__(self, joblist, asynchronous=False) -> None:
        self.jobs = joblist
        self.asynchronous = asynchronous
        if self.asynchronous:
            self.sched = BackgroundScheduler(timezone='Asia/Shanghai')
        else:
            self.sched = BlockingScheduler(timezone='Asia/Shanghai')
        self._check()
        self._add_jobs()

    def _check(self):
        def raise_display():
            raise CrontabError(
                "joblist like [ {'crontab':str,'func':FunctionType,'args':tuple}, ... ]")

        k = ['func', 'args', 'crontab']
        for job in self.jobs:
            for i in k:
                if i not in job.keys():
                    raise_display()
                if type(job['crontab']) is not str:
                    raise_display()
                if not isinstance(job['func'], types.FunctionType):
                    raise_display()
                if type(job.get('args', tuple)) is not tuple:
                    raise_display()
        return 0

    def _add_jobs(self):
        for job in self.jobs:
            func = job["func"]
            crontab = job["crontab"]
            args = job.get("args")
            self._add_job(func, crontab, args)

    def _add_job(self, func, crontab, args=None):
        cron = MyCronTrigger.from_crontab(crontab)
        self.sched.add_job(func, cron, args=args)

    def start(self):
        if self.asynchronous:
            self.sched._daemon = True
        self.sched.start()


if __name__ == '__main__':
    def p(*args):
        print(args, datetime.datetime.now())
        print('====\n')


    joblist = [
        {'crontab': '0/1 * * * *', 'func': p, 'args': ('job1',)},
        {'crontab': '0 */2 * * * ?', 'func': p, 'args': ('job2',)},
        {'crontab': '0 */3 * * * ? *', 'func': p, 'args': ('job3',)},
    ]

    # sync
    crontab1 = Crontab(joblist)
    crontab1.start()

    # async
    # import time
    # crontab2 = Crontab(joblist, True)
    # crontab2.start()
    # time.sleep(999999)



'''
Author: hexu
Date: 2021-10-14 14:54:05
LastEditTime: 2023-05-25 14:49:05
LastEditors: Hexu
Description: 本地csv数据集
FilePath: /iw-algo-fx/intelliw/datasets/datasource_local_csv.py
'''
import time
from intelliw.utils.exception import DatasetException
from intelliw.datasets.datasource_base import AbstractDataSource, DataSourceReaderException, DataSourceType, \
    DatasetSelector
from intelliw.utils.global_val import gl
from intelliw.utils.logger import _get_framework_logger
from intelliw.utils.spark_process import Engine

logger = _get_framework_logger()
engine = Engine()


class DataSourceLocalCsv(AbstractDataSource):
    """
    本地 csv 文件数据源
    """

    def __init__(self, file_path):
        self.file_path = file_path
        self.__total = -1
        self.__meta = []

    def total(self):
        if self.__total >= 0:
            return self.__total
        logger.info("start read files: %s", self.file_path)
        start_time = time.time()
        csv = engine.csv(self.file_path).read()
        self.__total = csv.index.size
        if self.__total == 0:
            raise DataSourceReaderException('csv 文件为空')
        logger.info("total: %s, load time: %s",
                    self.__total, time.time() - start_time)
        return self.__total

    def reader(self, pagesize=10000, offset=0, limit=0, transform_function=None, dataset_type='train_set'):
        return self.__DataSourceReaderLocalCsv(self.file_path, pagesize, offset, limit, transform_function)

    def get_meta(self):
        if self.__meta is not None and len(self.__meta) > 0:
            return self.__meta
        return []

    class __DataSourceReaderLocalCsv:
        """
        读取本地 csv
        """

        def __init__(self, file_path, pagesize=10000, offset=0, limit=0, transform_function=None):
            self.pagesize = pagesize
            self.is_stop = False
            self.file_path = file_path
            self.limit = limit
            self.meta = []  # csv 元数据
            self.read_count = 0
            self._init_meta()
            self.reader = self.__Reader(file_path, offset)
            self.transform_function = transform_function

        def get_meta(self):
            """获取元数据"""
            return self.meta

        def reset(self, fields: list=None, page_size: int=None):
            """重置读取字段和分页大小"""
            raise NotImplementedError("reset method not implemented")


        @property
        def iterable(self):
            return True

        def __iter__(self):
            return self

        def __next__(self):
            if self.is_stop:
                raise StopIteration
            try:
                data = self._read_batch()
                return_data = {'result': data, 'meta': self.meta}
                if self.transform_function is not None:
                    return_data = self.transform_function(return_data)
                return return_data
            except Exception as e:
                logger.exception("csv 文件读取错误")
                self.reader.close()
                raise DataSourceReaderException('csv 文件读取失败') from e

        def _read_batch(self):
            if self.limit > 0 and self.read_count + self.pagesize >= self.limit:
                batch_amount = self.limit - self.read_count
                self.is_stop = True
            else:
                batch_amount = self.pagesize
            batch = self.reader.read(batch_amount)
            self.read_count += batch_amount
            if self.reader.is_finish:
                self.is_stop = True
            return batch

        def _init_meta(self):
            csv = engine.csv(self.file_path).read(nrows=0)
            self.meta = [{'code': i} for i in csv.columns.values.tolist()]

            # TODO 很恶心的逻辑 后续看看怎么干掉
            dataset_idx = gl.get("dataset_idx", 0)
            target_metadata = gl.target_metadata
            if isinstance(target_metadata, list) and len(target_metadata) > dataset_idx:
                target_info: dict = target_metadata[dataset_idx]
                target_col: int = target_info["target_col"]
                if target_col > len(self.meta):
                    raise DatasetException(
                        f"target column index:{target_col}, There are {len(self.meta)} "
                        f"columns in the source csv, index out of range")
                self.meta[target_col] = {'code': target_info["target_name"]}

        class __Reader:
            def __init__(self, file_path, offset=0):
                self.line_offset = offset
                self.is_finish = False
                self.file_path = file_path

            def close(self):
                del self.csv

            def reset(self, fields: list=None, page_size: int=None):
                self.line_offset = 0
                self.pagesize = page_size
                self.csv = engine.csv(self.file_path).read(
                    skiprows=self.line_offset, nrows=page_size)
                self.csv.columns = fields

            def read(self, amount=1):
                try:
                    self.csv = engine.csv(self.file_path).read(
                        skiprows=self.line_offset, nrows=amount)
                except DatasetException:
                    if self.line_offset == 0:
                        msg = 'csv 文件为空'
                    else:
                        msg = '超出范围的 offset {}'.format(self.line_offset)
                    raise DataSourceReaderException(msg)

                batch = self.csv.values.tolist()
                if len(batch) < amount:
                    self.is_finish = True
                    self.line_offset += len(batch)
                    return batch
                self.line_offset += amount
                return batch


DatasetSelector.register_func(DataSourceType.LOCAL_CSV, DataSourceLocalCsv, {
    "file_path": "CSV_PATH"})



import os
import time
import json
from functools import wraps
import logging
import zipfile
from typing import Any, Callable, Dict, Optional

from intelliw.utils.logger import _get_framework_logger

logger = _get_framework_logger()


def model_cache(timeout: int = 3600):
    """
    模型缓存装饰器
    缓存模型加载结果，减少重复加载

    :param timeout: 缓存超时时间(秒)
    :return: 装饰器函数
    """
    cache = {}
    last_updated = 0

    def decorator(func: Callable) -> Callable:
        @wraps(func)
        def wrapper(*args: Any, **kwargs: Any) -> Any:
            nonlocal last_updated
            current_time = time.time()

            # 检查缓存是否过期
            if 'model' in cache and current_time - last_updated < timeout:
                logger.debug("Using cached model")
                return cache['model']

            # 加载新模型
            model = func(*args, **kwargs)

            # 更新缓存
            cache['model'] = model
            last_updated = current_time
            logger.debug("Model cache updated")

            return model

        return wrapper

    return decorator


def validate_model(func: Callable) -> Callable:
    """
    模型验证装饰器
    在加载模型后验证模型有效性

    :param func: 被装饰的函数
    :return: 装饰器函数
    """
    @wraps(func)
    def wrapper(self, *args: Any, **kwargs: Any) -> Any:
        model = func(self, *args, **kwargs)

        # 验证模型
        # if hasattr(self, 'model_loader') and hasattr(self.model_loader, 'validate'):
        #     if not self.model_loader.validate(model):
        #         raise ValueError("Invalid model loaded")

        return model

    return wrapper


def model_update_notification(func: Callable) -> Callable:
    """
    模型更新通知装饰器
    当模型更新时发送通知

    :param func: 被装饰的函数
    :return: 装饰器函数
    """
    @wraps(func)
    def wrapper(self, *args: Any, **kwargs: Any) -> Any:
        result = func(self, *args, **kwargs)

        # 如果有更新，发送通知
        if result and isinstance(result, dict) and 'version' in result:
            logger.info(f"Model updated to version: {result['version']}")
            # 这里可以添加通知逻辑，如发送消息到消息队列等

        return result

    return wrapper


def model_save(func):
    return func

def train_info_report(func):
    return func

def val_info_report(func):
    return func


def zip_model(model_path: str, output_path: str) -> bool:
    """
    压缩模型文件

    :param model_path: 模型文件路径
    :param output_path: 输出zip文件路径
    :return: 是否压缩成功
    """
    try:
        with zipfile.ZipFile(output_path, 'w', zipfile.ZIP_DEFLATED) as zipf:
            if os.path.isdir(model_path):
                for root, _, files in os.walk(model_path):
                    for file in files:
                        file_path = os.path.join(root, file)
                        arcname = os.path.relpath(file_path, os.path.dirname(model_path))
                        zipf.write(file_path, arcname)
            else:
                zipf.write(model_path, os.path.basename(model_path))
        logger.info(f"Model compressed to {output_path}")
        return True
    except Exception as e:
        logger.error(f"Error compressing model: {str(e)}")
        return False


'''
Author: Hexu
Date: 2022-08-16 10:23:16
LastEditors: Hexu
LastEditTime: 2022-09-27 10:06:25
FilePath: /iw-algo-fx/intelliw/utils/database/connection.py
Description:

数据库连接层，SQLAlchemy的基础上wrap了一层接口，
负责自动管理SQLAlchemy的engine/session/scoped_session的创建，注册与销毁。

一个数据库uri，对应一个engine，一个sessionmaker工厂类，一个scoped_session工厂类
若要获取session/scoped_session对象执行数据库相关的操作，则：

引擎池由该模块进行维护。采用注册机制，通过user + password + host + db构造uri
然后获取一个SQLAlchemy数据库引擎，如果没有引擎则自动创建再返回。
其中，uri为'mysql+pymysql://{user}:{passwd}@{host}'的形式。
'''

import warnings

from sqlalchemy import create_engine as __create_engine
from sqlalchemy.engine import Engine
from sqlalchemy.orm import sessionmaker, scoped_session
from sqlalchemy.orm.session import Session

warnings.simplefilter('ignore')

# 引擎池，uri -> engine obj
__engines = {}

# sessionmakers工厂类池，engine obj -> sessionmakers
__SessionFactories = {}

# scoped_session工厂类池，engine obj -> scoped_session
__ScopedSessionFactories = {}


def get_engine(driver=None,
               user=None, password=None, host=None, db=None, uri=None,
               charset='utf8', use_cache=True, echo=False) -> Engine:
    """

    Args:
        driver(str): 驱动, e.g.: mysql+pymysql / postgresql+psycopg2 / oracle+cx_oracle
        user(str): 用户名
        password(str): 密码
        host(str): ip或者'localhost'
        db(str or None): 数据库名称
        charset(str): 字符集
        use_cache(bool): 为True时，用完engine记得dispose
        echo(bool): 是否显示sql

    Returns:

    """
    if uri is None:
        uri = '{driver}://{user}:{password}@{host}/'
        uri = uri.format(driver=driver, user=user,
                         password=password, host=host)
        uri += '{}'.format(db) if db else ''
        uri += '?charset={}'.format(charset) if charset else ''
    if use_cache:
        if uri not in __engines:
            __engines[uri] = __create_engine(uri, echo=echo)
        return __engines[uri]
    else:
        return __create_engine(uri, echo=echo)


def get_scoped_session(engine) -> Session:
    """Get a Thread-Local Storage SQLAlchemy session object"""
    return get_scoped_session_factory(engine)()


def get_session(engine) -> Session:
    """Get a non thread-safe SQLAlchemy session"""
    warnings.warn(
        '正在使用一个非线程安全的Session！推荐使用get_scoped_session(engine)函数来获取session。')
    return get_session_factory(engine)()


def get_session_factory(engine):
    if not __SessionFactories.get(engine):
        __SessionFactories[engine] = sessionmaker(bind=engine)
    return __SessionFactories[engine]


def get_scoped_session_factory(engine):
    if not __ScopedSessionFactories.get(engine):
        if not __SessionFactories.get(engine):
            __SessionFactories[engine] = sessionmaker(bind=engine)
        __ScopedSessionFactories[engine] = scoped_session(
            __SessionFactories[engine])
    return __ScopedSessionFactories[engine]



#!/usr/bin/env python
# coding: utf-8
import datetime
import inspect
import logging
import math
import string
import traceback
import types
import json
import os
import zipfile
from random import Random
import ctypes
import logging

from intelliw.config import config

logger = logging.getLogger()
random = Random()

try:
    import numpy as np

    np.set_printoptions(threshold=np.inf)
    has_np = True
except ImportError:
    has_np = False


def prepare_algorithm_parameters(alg_ps, value=None):
    value = value if value is not None else {}
    options = {}
    if alg_ps is not None:
        for k, v in enumerate(alg_ps):
            key = v['key'] if 'key' in v else None
            option = v['option'] if 'option' in v else None
            val = v['val'] if 'val' in v else None
            if key is not None:
                value[key] = val
                options[key] = option
                if option is not None and type(option) == dict and 'type' in option:
                    try:
                        val = to_type(val, option['type'])
                    except:
                        pass
                    value[key] = val
    return value, options


def prepare_model_parameters(model_ps, value=None, option=None):
    value = value if value is not None else {}
    if model_ps is not None:
        assert (type(model_ps) is dict)
        for k, v in model_ps.items():
            value[k] = v
            if option is not None and type(option) == dict and option.get(k) is not None and type(
                    option[k]) == dict and 'type' in option[k]:
                try:
                    v = to_type(v, option[k]['type'])
                except:
                    pass
                value[k] = v
    return value


def to_type(val, t: str):
    t = t.lower()
    if t == 'string' or t == 'str':
        return str(val)
    if t == 'enum':
        return str(val)
    if t == 'integer' or t == 'int':
        return int(val)
    if t == 'double' or t == 'float':
        return float(val)
    if t == 'boolean' or t == 'bool':
        if isinstance(val, bool):
            return val
        return val == 1 or val == '1' or val == 'True' or val == 'true'
    if t == 'array(int)' or t == 'array(integer)':
        return list(map(int, val.split(",")))
    if t == 'array(string)' or t == 'array(str)':
        return val.split(",")
    if t == 'array(array(int))' or t == 'array(array(integer))':
        return [list(map(int, filter(lambda n: n != "", val.split(","))))]
    return val


def get_json_encoder():
    global has_np
    if has_np:
        class JsonEncoder(json.JSONEncoder):
            """Convert numpy classes to JSON serializable objects."""

            def default(self, obj):
                if isinstance(obj, (np.integer, np.floating, np.bool_)):
                    return obj.item()
                elif isinstance(obj, np.ndarray):
                    return obj.tolist()
                elif isinstance(obj, (datetime.datetime, datetime.timedelta)):
                    return obj.__str__()
                elif isinstance(obj, types.GeneratorType):
                    try:
                        return list(obj)
                    except Exception:
                        return {"__generator__": str(obj)}
                elif isinstance(obj, types.AsyncGeneratorType):
                    return self._extract_async_generator_info(obj)
                elif isinstance(obj, (set, frozenset)):
                    return list(obj)
                elif isinstance(obj, bytes):
                    return obj.decode('utf-8', errors='replace')
                elif isinstance(obj, complex):
                    return {"real": obj.real, "imag": obj.imag}
                elif obj.__class__.__name__ == 'StreamingResponse':
                    return self._extract_streaming_response_info(obj)
                else:
                    return super(JsonEncoder, self).default(obj)

            def _extract_async_generator_info(self, obj):
                info = {"__async_generator__": True}
                try:
                    if hasattr(obj, 'ag_code') and obj.ag_code:
                        info["func_name"] = getattr(obj.ag_code, 'co_name', None)
                        info["file_path"] = getattr(obj.ag_code, 'co_filename', None)
                        info["first_line"] = getattr(obj.ag_code, 'co_firstlineno', None)
                except Exception:
                    pass
                try:
                    info["running"] = obj.ag_running
                except Exception:
                    pass
                return info

            def _extract_streaming_response_info(self, obj):
                info = {"__streaming_response__": True}
                try:
                    info["media_type"] = getattr(obj, 'media_type', None)
                    info["status_code"] = getattr(obj, 'status_code', None)
                except Exception:
                    pass
                return info

        return JsonEncoder
    else:
        class JsonEncoder(json.JSONEncoder):
            """Convert common non-JSON-serializable types to JSON serializable objects."""

            def default(self, obj):
                if isinstance(obj, (datetime.datetime, datetime.timedelta)):
                    return obj.__str__()
                elif isinstance(obj, types.GeneratorType):
                    try:
                        return list(obj)
                    except Exception:
                        return {"__generator__": str(obj)}
                elif isinstance(obj, types.AsyncGeneratorType):
                    return self._extract_async_generator_info(obj)
                elif isinstance(obj, (set, frozenset)):
                    return list(obj)
                elif isinstance(obj, bytes):
                    return obj.decode('utf-8', errors='replace')
                elif isinstance(obj, complex):
                    return {"real": obj.real, "imag": obj.imag}
                elif obj.__class__.__name__ == 'StreamingResponse':
                    return self._extract_streaming_response_info(obj)
                else:
                    return super(JsonEncoder, self).default(obj)

            def _extract_async_generator_info(self, obj):
                info = {"__async_generator__": True}
                try:
                    if hasattr(obj, 'ag_code') and obj.ag_code:
                        info["func_name"] = getattr(obj.ag_code, 'co_name', None)
                        info["file_path"] = getattr(obj.ag_code, 'co_filename', None)
                        info["first_line"] = getattr(obj.ag_code, 'co_firstlineno', None)
                except Exception:
                    pass
                try:
                    info["running"] = obj.ag_running
                except Exception:
                    pass
                return info

            def _extract_streaming_response_info(self, obj):
                info = {"__streaming_response__": True}
                try:
                    info["media_type"] = getattr(obj, 'media_type', None)
                    info["status_code"] = getattr(obj, 'status_code', None)
                except Exception:
                    pass
                return info

        return JsonEncoder


def generate_random_str(length=16):
    return ''.join(random.sample(string.ascii_letters + string.digits, length))


def import_code(code, mod=None, name=None):
    if mod is None:
        # create blank module
        name = name if name is not None else generate_random_str()
        mod = types.ModuleType(name)
    exec(code, mod.__dict__)
    return mod


def get_first_element(mod):
    for element_name in dir(mod):
        element = getattr(mod, element_name)
        if inspect.isclass(element):
            return element()
        elif inspect.isfunction(element):
            return element


def is_empty_function(f):
    """Returns true if f is an empty function."""

    def empty_func():
        pass

    def empty_func_with_docstring():
        """Empty function with docstring."""
        pass

    def empty_lambda(): return None

    def empty_lambda_with_docstring(): return None

    empty_lambda_with_docstring.__doc__ = """Empty function with docstring."""

    def constants(f):
        """Return a tuple containing all the constants of a function without:
            * docstring
        """
        return tuple(
            x
            for x in f.__code__.co_consts
            if x != f.__doc__
        )

    return (
            f.__code__.co_code == empty_func.__code__.co_code and
            constants(f) == constants(empty_func)
    ) or (
            f.__code__.co_code == empty_func_with_docstring.__code__.co_code and
            constants(f) == constants(empty_func_with_docstring)
    ) or (
            f.__code__.co_code == empty_lambda.__code__.co_code and
            constants(f) == constants(empty_lambda)
    ) or (
            f.__code__.co_code == empty_lambda_with_docstring.__code__.co_code and
            constants(f) == constants(empty_lambda_with_docstring)
    )


def unzip_file(zipfilename: str, unziptodir=None):
    unziptodir = zipfilename.rstrip(".zip")
    if not os.path.exists(unziptodir):
        os.mkdir(unziptodir, 0o755)
    if zipfile.is_zipfile(zipfilename):
        fz = zipfile.ZipFile(zipfilename, 'r')
        for file in fz.namelist():
            fz.extract(file, unziptodir)
    else:
        raise FileExistsError('This is not zip')
    return unziptodir


def get_worker_count(cpus, threads, is_multi_precess):
    worker, thread = 1, number_of_threads(cpus)
    if is_multi_precess:
        worker = number_of_workers(config.CPU_COUNT)

    # 自定义线程数
    if threads:
        try:
            thread = int(threads)
            if worker > 1:
                thread = math.floor(thread / worker)
            thread = min(128, max(1, thread))
        except Exception as e:
            traceback.print_exc()

    return worker, thread


def number_of_workers(core: str = ''):
    core = str(core).lower()
    num_core = 1
    try:
        if core:
            if 'm' in core:
                num_core = 1
            elif core.isdigit():
                num_core = int(core)
            else:
                num_core = math.ceil(float(core))
    except:
        pass
    # docker容器里取的是宿主机的
    # return (multiprocessing.cpu_count() * 2) + 1
    # 最优是 2*core+1
    # 但是服务器和算法程序都太脆弱了， 所以减少点
    # return (2 * num_core) + 1
    return min(8, num_core)


def number_of_threads(core: str = ''):
    core = str(core).lower()
    num_core = 1
    try:
        if core:
            if 'm' in core.lower():
                num_core = 1
            elif core.isdigit():
                num_core = int(core)
            else:
                num_core = math.ceil(float(core))
    except:
        pass
    return min(64, 2 * num_core + 1)


def stop_thread(tid, exctype):
    """raises the exception, performs cleanup if needed"""
    tid = ctypes.c_long(tid)
    if not inspect.isclass(exctype):
        exctype = type(exctype)
    res = ctypes.pythonapi.PyThreadState_SetAsyncExc(tid, ctypes.py_object(exctype))
    if res == 0:
        logger.info("invalid thread id")
        # raise ValueError("invalid thread id")
    elif res != 1:
        # """if it returns a number greater than one, you're in trouble,
        # and you should call it again with exc=NULL to revert the effect"""
        ctypes.pythonapi.PyThreadState_SetAsyncExc(tid, None)
        raise SystemError("PyThreadState_SetAsyncExc failed")


def gen_random_str(length=4):
    import random
    import string
    return ''.join(random.choices(string.ascii_letters + string.digits, k=length))


def string_to_function(func_str, func_name):
    """
    将包含函数定义的字符串转换为实际的函数。

    :param func_str: 包含函数定义的字符串
    :param func_name: 函数的名称
    :return: 动态定义的函数
    """
    # 创建一个字典作为命名空间
    local_namespace = {}

    # 尝试执行字符串代码
    try:
        exec(func_str, globals(), local_namespace)
    except Exception as e:
        print(f"Error executing function string: {e}")
        return None

    # 从命名空间中提取函数
    func = local_namespace.get(func_name)

    if func is None:
        print(f"Function {func_name} is not defined in the provided string.")
        return None

    return func

def use_ipv6_by_env():
    '''
    是否使用ipv6
    :return:
    '''
    return os.environ.get('USE_IPV6', 'false').lower() == 'true' or os.environ.get('useIPv6', 'false').lower() == 'true'

def parse_host_port_from_netloc(netloc):
    '''
    解析host和port
    :param netloc: host:port
    :return: host, port
    '''
    # 处理IPv6地址格式 [ipv6]:port 或 ipv4:port
    if ']:' in netloc and netloc.startswith('['):
        # IPv6格式: [ipv6地址]:端口
        bracket_end = netloc.find(']')
        host = netloc[1:bracket_end]  # 不包含方括号
        port_part = netloc[bracket_end + 2:]  # 跳过]: 
        port = int(port_part) if port_part.isdigit() else ''
    else:
        # IPv4或域名格式: host:port
        tmp_list = netloc.split(':')
        if len(tmp_list) >= 2 and tmp_list[-1].isdigit():
            host = ':'.join(tmp_list[:-1])
            port = int(tmp_list[-1]) if tmp_list[-1].isdigit() else ''
        else:
            host = netloc
            port = ''
    
    return host, port

DB_NUM_TYPE = [
    "TINYINT",
    "SMALLINT",
    "MEDIUMIN",
    "INT",
    "INTEGER",
    "BIGINT",
    "NUMERIC",
    "FLOAT",
    "DOUBLE",
    "DECIMAL",
]



import os
import logging


logger = logging.getLogger('env_file')


def read_env_file():
    try:
        env_file = os.environ.get('ENV_FILE')
        if not env_file:
            logger.warning("未找到ENV_FILE配置变量, 返回")
            return
        # 判断文件列表中的文件都存在 否则返回错误信息
        if not os.path.exists(env_file):
            logger.warning(f"未找到配置文件 {env_file}, 返回")
            return

        with open(env_file, 'r') as f:
            for line in f.readlines():
                key, value = line.strip().split('=', 1)
                os.environ[key] = value
    except Exception as e:
        logger.error("ENV_FILE配置文件读取失败", e)


#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Time    : 2024/3/26 15:06
# @Author  : HEXIN
# @File    : cGroup.py.py
# @Description :

# refer from test_004
from pathlib import Path
import subprocess
import argparse
from multiprocessing import Process, cpu_count
import os

CGROUP_ROOT = '/sys/fs/cgroup'
PROCESS_CONTROL_PATH = f'{CGROUP_ROOT}/cpuset/'
GROUP_NAME_TEMPLATE = 'hTest_'


def _argparse():
    """
    get input arguments
    :return: argument object
    """
    parser = argparse.ArgumentParser(description="hexin test")
    parser.add_argument('--total_process_number', action='store', dest='totalProcessNumber', default="2",
                        help='Total number of processes')
    parser.add_argument('--total_thread_number_per_process', action='store', dest='totalThreadNumberPerProcess',
                        default="2", help='Total number of theads per process')
    parser.add_argument('--mprocessing_in_thread', action='store_true', dest='is_mprocessing_in_thread',
                        help='Run multiple process in main thread')

    return parser.parse_args()


def cgroup_maker(group_number: int = 2, thread_pre_group: int = 2, is_run_mp_in_thread: bool = False):
    """
    make cgroup
    """

    def _get_group_list() -> list:

        root: Path = Path(PROCESS_CONTROL_PATH)

        custom_group_list = []
        for _p in root.iterdir():

            if _p.is_dir():
                # TODO debug
                # print(_p)

                # remove parent path
                _p_name = _p.name.replace(PROCESS_CONTROL_PATH, '')

                if _p_name.startswith(GROUP_NAME_TEMPLATE):
                    # filter via name template
                    custom_group_list.append(_p_name)

        # TODO debug
        for _g in custom_group_list:
            print(_g)

        return custom_group_list

    def _remove_exist_groups(group_list: list = []):

        # remove exist groups
        for _g in group_list:
            wrk_path = f'{PROCESS_CONTROL_PATH}{_g}'
            print(f'delete {wrk_path}')

            result = subprocess.run(['rmdir', wrk_path], capture_output=True, text=True)

            if not (result.stderr is None or '' == result.stderr):
                print(result.stderr)

    def _create_new_groups():
        for i in range(group_number):
            # create new group
            _new_group_path = f'{PROCESS_CONTROL_PATH}{GROUP_NAME_TEMPLATE}{i}'
            print(f'creating {_new_group_path}')

            # create folder
            result = subprocess.run(['mkdir', _new_group_path], capture_output=True, text=True)
            if not (result.stderr is None or '' == result.stderr):
                print(result.stderr)

            # specify CPU core
            with open(f'{_new_group_path}/cpuset.cpus', 'w') as f:
                f.write(f'{i * thread_pre_group}-{(i + 1) * thread_pre_group - 1}')
            # specify mem node
            with open(f'{_new_group_path}/cpuset.mems', 'w') as f:
                f.write('0')

            if not is_run_mp_in_thread:
                p = Process(target=deadloop)
                p.start()
                # specify pid
                with open(f'{_new_group_path}/tasks', 'w') as f:
                    f.write(str(p.pid))
            else:
                p = Process(target=sub_process, args=(i, thread_pre_group))
                p.start()

                # TODO debug
                print(f'start sub-process with group id = {i}, pid = {p.pid}')

                # specify pid
                with open(f'{_new_group_path}/cgroup.procs', 'w') as f:
                    f.write(str(p.pid))

    # get exist group list
    exist_group_name_list = _get_group_list()
    # remove old ones
    _remove_exist_groups(exist_group_name_list)
    # create new groups with default template
    _create_new_groups()


def deadloop():
    while True:
        pass


def sub_process(current_group_number: int = 0, thread_pre_group: int = 2):
    _sub_pid = os.getpid()

    for i in range(thread_pre_group):
        p = Process(target=deadloop)
        p.start()
        print(f'Run sub-process in group {current_group_number}, parent pid = {_sub_pid}, local pid = {p.pid}')


if __name__ == '__main__':

    args = _argparse()

    # total process numer
    total_process_number = int(args.totalProcessNumber)
    # total thread numer per process
    total_thread_number_per_process = int(args.totalThreadNumberPerProcess)
    # run multiple process in main thread
    is_mprocessing_in_thread = args.is_mprocessing_in_thread

    # verify
    if total_process_number * total_thread_number_per_process > cpu_count():
        raise ValueError(f'invalid process and thread number')

    cgroup_maker(total_process_number, total_thread_number_per_process, is_mprocessing_in_thread)
